"""
***模块的父类模板*** \n
"""


import torch.nn as nn
from typing import Any, Dict, Tuple, List
from .const import ModuleKey, uuid_t, ModuleMode
from .error import ModelError, ModelWarning
from .node import ModuleNode


class ModuleTemp(ModuleNode):
    """模块的通用模板父类"""

    # 打印时的留空
    _blank = 4

    def __init__(self, nodes: Dict[uuid_t, ModuleNode] | None = None,
                 info: Dict[uuid_t, Any] | None = None) -> None:
        """
        创建结构时的初始化 \n

        :param nodes: 包含所有ModuleNode格式的module字典 \n
        :param info: 信息表 \n
        """
        # _nodes为存储所有key对应部分的容器
        # _model为仅存储所有nn.Module模型的nn容器，每次刷新modules结构都会重建_models
        # _key2index记录了从dict到list的序号映射
        # _mask记录了是否屏蔽其中的某个部分
        # self是可运行的模型
        for node in nodes.values():
            self.check_node(node)
        self._nodes = ModuleKey.dict_(nodes)
        self._mask_dict: Dict[uuid_t, bool] = dict()
        for key in ModuleKey.iter_():
            self._mask_dict[key] = False
        self._mask_list: List[bool] | None = None
        model, self._key2index = self._build_model()
        super().__init__(model, info)

    def _build_model(self) -> Tuple[nn.ModuleList, Dict[uuid_t, int]]:
        """由module构建model和key2index，同时检查类型正确性，刷新掩码表"""
        index = 0
        key2index = dict()
        model = nn.ModuleList()
        self._mask_list = list()
        for key in ModuleKey.iter_():
            node = self._nodes[key]
            if node is not None:
                model.append(node)
                self._mask_list.append(self._mask_dict[key])
                key2index[key] = index
                index += 1
        return model, key2index

    def to_mask(self, key: uuid_t, mask: bool) -> None:
        """改变屏蔽状态，`mask` 为 `True` 时进行屏蔽，否则解屏蔽"""
        self._mask_dict[key] = mask
        self._mask_list[self._key2index[key]] = mask

    def module(self, key: uuid_t) -> ModuleNode:
        """返回模块"""
        return self._nodes[key]

    def print_model(self, blank: int = 0) -> None:
        """打印模型构成"""
        blank_str = ' ' * blank
        print(f"{blank_str}ModuleTemp：{self.__class__.__name__}")
        blank += self._blank
        blank_str += ' ' * self._blank
        for key in ModuleKey.iter_():
            module = self._nodes[key]
            print(f"{blank_str}{ModuleKey.get_name_(key)}：", end='')
            if module is None:
                print("None")
            else:
                print()
                module.print_model(self._blank + blank)

    def forward(self, x):
        """前向传播，若model为空的Sequential则会报错"""
        for i in range(len(self._model)):
            if not self._mask_list[i]:
                x = self._model[i](x)
        return x

    def update(self, key: uuid_t, node: ModuleNode | None):
        """更新模块，key指定模块号"""
        ModuleKey.check_in_(key, all_or_index=False)
        self.check_node(node)
        # 若删除模型结构
        pre_node = self._nodes[key]
        self._nodes[key] = node
        if node is None and pre_node is not None:
            del self._model[self._key2index[key]]
            del self._key2index[key]
        # 若添加模型结构，nn.ModuleList不支持insert操作，因此每次更新时都会重构该这一部分
        elif node is not None and pre_node is not None:
            self._model, self._key2index = self._build_model()
        # 若替换模型结构
        else:
            pass

    def load_weights(self, weights_src: str | Any, is_path: bool = False, key: uuid_t = ModuleKey.all__) -> None:
        """
        根据is_path，选择从路径或从内存中加载指定部分的模块参数 \n
        `key` 指定模块号
        """
        ModuleKey.check_in_(key, all_or_index=True)
        if key == ModuleKey.all__:
            load_weights = super().load_weights
        else:
            load_weights = self._nodes[key].load_weights
        return load_weights(weights_src, is_path)

    def save_weights(self, path: str | None = None, key: uuid_t = ModuleKey.all__) -> Any:
        """
        根据path，选择加载指定部分的模块参数到路径或从内存中 \n
        `key` 指定模块号
        """
        ModuleKey.check_in_(key, all_or_index=True)
        if key == ModuleKey.all__:
            save_weights = super().save_weights
        else:
            save_weights = self._nodes[key].save_weights
        return save_weights(path)



"""
***训练器***
"""
import os.path

import torch.nn as nn
import torch
from typing import Dict, Any
from .const import uuid_t, TrainerInfo
from ..temp import ModuleNode, ModuleMode
from .monitor import TrainerMonitor
from .optimizer import OptimTemp
from .criterion import CriterionTemp
from .evaluator import EvaluatorTemp
from .dataloader import DataLoaderTemp


class TrainerTemp:
    """训练器模板"""

    def __init__(self, model: ModuleNode,
                 optim: OptimTemp,
                 criterion: CriterionTemp,
                 dataloader: DataLoaderTemp,
                 monitor: TrainerMonitor,
                 evaluator: EvaluatorTemp,
                 info: Dict[uuid_t, Any] | None = None):
        """
        初始化容器，基准训练器本身不负责具体实现，仅负责调度各个子容器并协调配置 \n

        :param model: 模型 \n
        :param optim: 优化器 \n
        :param criterion: 损失函数计算 \n
        :param dataloader: 数据管理器 \n
        :param monitor: 监视器与日志记录器 \n
        :param evaluator: 评估器 \n
        :param info: 配置信息 \n

        ..todo::
            定义训练截止条件
        """
        # 检查信息表
        info = TrainerInfo.dict_(info)
        # 容器赋值
        self._model = model
        self._dataloader, self._optim, self._criterion = dataloader, optim, criterion
        self._monitor, self._evaluator = monitor, evaluator
        self._info = info
        # 交叉配置各部分
        self._model.to(self._info[TrainerInfo.device])
        self._monitor.init(self._info)
        self._optim.init(self._model)
        self._criterion.init(self._model)
        self._dataloader.init(self._model.info)
        # 最后配置评估器
        self._evaluator.init(self._model, self._dataloader, self._criterion)

    def train(self):
        """按照配置方式进行训练"""
        self._monitor.start()
        max_effect = 0.0
        now_path, old_path = None, None
        for epoch in range(self._info[TrainerInfo.max_epochs]):
            self._monitor.new_epoch(epoch)
            train_loss, sample_num = 0.0, 0
            self._model.to(mode=ModuleMode.train)
            for i, data in enumerate(self._dataloader.train):
                inputs, labels = data
                self._optim.zero_grad()
                outputs = self._model(inputs)
                loss = self._criterion(outputs, labels)
                loss.backward()
                self._optim.step()
                train_loss += loss * inputs.size(0)
                sample_num += inputs.size(0)
            train_loss /= sample_num
            val_loss, val_effect = self._evaluator.evaluate()
            self._monitor.record_effect(epoch, train_loss, val_loss, val_effect)
            # todo：进行可配置的参数保存，这里直接实现保存最大值
            if val_effect > max_effect:
                max_effect = val_effect
                now_path = os.path.join(self._monitor.logdir, f"{epoch}: {max_effect}")
                self._model.save_weights(now_path)
                if old_path is not None:
                    os.remove(old_path)
                old_path = now_path
        self._monitor.end()

    def end(self):
        """
        结束训练器并销毁资源占用 \n
        .. todo::
            未实现
        """



"""
工具类
"""


import matlab


m_double = matlab.double
m_int16 = matlab.int16
m_int32 = matlab.int32
m_int64 = matlab.int64
m_uint8 = matlab.uint8
m_uint16 = matlab.uint16
m_uint32 = matlab.uint32
m_uint64 = matlab.uint64




"""
***对于图像容器的函数API***
"""


from .im import *


def import_image(path: str, style: uuid_t, is_cache: bool = False) -> ImStd:
    """
    从路径中导入图像 \n
    `style` 指定了导入方式 \n
    `is_cache` 指定了是否是从缓存中导入，期望类型由 `style` 指定，可以为 `auto__`  \n
    `scale`，如果从路径导入，一般为 `uint8`，若从 cache 导入则与存入情况一样 \n
    `freq`，只能从缓存中导入，因为一般图片无法有效保存频谱信息 \n
    如果以 `cv2.imread` 方式导入，默认指定 `cv2.IMREAD_ANYCOLOR`，否则灰度图也会有3个通道 \n
    """
    ImStyle.check_in_(style, all_or_index=True)
    check_path_readable(path)
    if is_cache:
        im: ImStd = load_data_from_pkl(path)
    else:
        if style == ImStyle.cv_ndarray:
            image = cv2.imread(path, cv2.IMREAD_ANYCOLOR)
        else:
            image = Image.open(path)
            if style == ImStyle.im_file:
                pass
            else:
                image = np.array(image)
                if style == ImStyle.im_ndarray:
                    pass
                elif style == ImStyle.im_tensor:
                    image = torch.tensor(image, device='cpu')
                    if len(image.shape) == 2:
                        image.unsqueeze_(0)
                else:
                    raise ImageStyleError(
                        f"ERROR: Image style {ImStyle.get_value_(style)} is not implemented an 'import' function!!!")
        im = ImStd(image, style)
    return im




"""
图像直方图
"""


from .file.image.im import *
import numpy as np
from PIL import Image
import matplotlib.pyplot as plt
import typing


def draw_histogram(histogram, save: None | str = None, show: bool = True) -> None:
    """保存或显示直方图"""
    if save is not None or show is True:
        plt.figure(figsize=(10, 5))
        plt.bar(range(256), histogram, color='blue')
        plt.title('Grayscale Histogram')
        plt.xlabel('Pixel Intensity')
        plt.ylabel('Frequency')
        if save is not None:
            plt.savefig(save)
        if show is True:
            plt.show()


def draw_cdf(raw_cdf: typing.Sequence, fla_cdf: typing.Sequence, ide_cdf: typing.Sequence,
             show: bool = True, save: None | str = None) -> None:
    """保存或显示cdf"""
    if save is not None or show is True:
        plt.figure()
        plt.plot(range(256), raw_cdf, label='raw cdf', color='b')
        plt.plot(range(256), fla_cdf, label='flatten cdf', color='r')
        plt.plot(range(256), ide_cdf, label='ideal cdf', color='g')
        plt.title('CDF Comparison')
        plt.xlabel('Pixel Intensity')
        plt.ylabel('Frequency')
        plt.legend()
        if save is not None:
            plt.savefig(save)
        if show is True:
            plt.show()


def convert_histogram(im: ImStd) -> np.ndarray:
    """提取图片各通道的灰度直方图"""
    # 确保图片是灰度图
    im = im.to_color(ImColor.gray).to_style(ImStyle.im_file)
    image = im.image
    histogram = np.array(image).astype('float')
    # 标准化直方图
    num_pixels = image.size[0] * image.size[1]
    histogram_nm = histogram / num_pixels
    return histogram_nm


def flatten_histogram(im: ImStd) -> tuple:
    """直方图均衡化"""
    # 初始化结构
    quan_size = 256
    raw_his = convert_histogram(im)
    ideal_cdf = np.linspace(0, 1, quan_size, dtype=float)
    raw_cdf = np.zeros(quan_size, dtype=float)
    fla_cdf = np.zeros(quan_size, dtype=float)
    map_sheet = np.zeros(quan_size, dtype=int)
    fla_his = np.zeros(quan_size)

    # 建立raw_cdf
    raw_cdf[0] = raw_his[0]
    for i in range(1, quan_size):
        raw_cdf[i] = raw_his[i] + raw_cdf[i - 1]

    # 遍历raw_his来产生fla_his
    j = 0   # j是fla_cdf指针
    for i in range(0, quan_size):
        # 处理饱和情况
        if j == quan_size:
            fla_his[j] += raw_his[i]
            map_sheet[i] = j
            fla_cdf[j] += raw_his[i]
            continue
        # 处理零情况
        if raw_his[i] == 0.0:
            if i >= 1:
                map_sheet[i] = map_sheet[i - 1]
            continue
        # 处理一般情况
        tmp_cdf = fla_cdf[j] + raw_his[i]
        n = j
        for n in range(j, quan_size):
            # 如果此时比理想值小，那么这就是映射的目标点，设置该点cdf和映射值，并退出循环
            # 这里使用的是优先填充cdf的策略
            # 第二个条件检查饱和情况（防止浮点数精度导致的误差）
            if tmp_cdf <= ideal_cdf[n] or n == quan_size - 1:
                fla_his[n] = raw_his[i]
                map_sheet[i] = n
                fla_cdf[n] = tmp_cdf
                break
            # 如果此时比理想值大，那么应继续寻找映射目标点，同时设置该点cdf
            else:
                fla_cdf[n] = fla_cdf[j]
        # 更新fla_cdf指针
        j = n

    im = im.to_style(ImStyle.im_ndarray).to_color(ImColor.gray)
    image = im.image
    for i in range(image.shape[0]):
        for j in range(image.shape[1]):
            image[i, j] = map_sheet[image[i, j]]
    image_fla = Image.fromarray(image)

    return image_fla, raw_his, fla_his, raw_cdf, fla_cdf, ideal_cdf













"""
***常用 `module_node`***
"""


from .module import nn, ModuleNode, ModuleTemp, ModuleKey, uuid_t, ModuleMode


class IdentityUnit(nn.Module):
    """恒等变换单元"""
    def __init__(self):
        super().__init__()

    def forward(self, x):
        """直接返回输入"""
        return x


class IdentityModel(ModuleNode):
    """恒等变换模块"""
    def __init__(self):
        super().__init__(IdentityUnit())


class ModuleTemp_MaskHead(ModuleTemp):
    """根据 (Train, Val) / Predict 模式决定是否掩码头部"""
    def _to_mode(self, mode: uuid_t) -> None:
        """在由预测转换为评估或训练时掩码头部"""
        super()._to_mode(mode)
        if mode == ModuleMode.predict:
            self.to_mask(ModuleKey.head, False)
        else:
            self.to_mask(ModuleKey.head, True)



"""
使用示例，本质上是作者做过的各项目与任务中的小规模部分的代码的复制粘贴 \n
示例中的代码会指定编写时的套件版本号，可能存在过时风险，并且对于框架的使用用的是绝对地址包导入 \n
因此，demo中的代码是不存在高度通用性的，但是可以提供包的使用的具体应用场景 \n
同时，如果应用场景碰巧与当前开发场景相同，可以考虑导入其中的部分代码，或直接将示例拷贝为当前开发场景的模板代码，在此基础上进行二次开发
"""



"""
***torchvision库中的数据集的标准化封装*** \n
`tcv` 指的是 `torchvision`
"""


from .dataset import DatasetTemp, _DatasetTemp, uuid_t, DatasetInfo, ModuleInfo, Dataset
from typing import Dict, Any
import torch
import torchvision.datasets as datasets
from hagike.utils.cache import MemCacher


class _TcvDataset(_DatasetTemp):
    """
    二次封装tcv数据集本身，转换独热码并使用内存缓存
    """
    def __init__(self, dataset: Dataset, para: Dict[str, Any]) -> None:
        """
        初始化 \n
        若启用缓存器则需要在 `para` 中定义 `cache_size` 项
        """
        super().__init__(dataset, para)
        if 'cache_size' in self._para:
            self._getter = MemCacher(self._getitem, self._para['cache_size'])
        else:
            self._getter = self._getitem

    def __getitem__(self, index):
        """获取数据"""
        return self._getter(index)

    def _getitem(self, index):
        """数据转换"""
        # torchvision的分类数据集中默认是非独热码，这里需要转换为独热码
        inputs, labels = super().__getitem__(index)
        one_hot = torch.zeros(self._para['num_classes'],
                              device=self._para[ModuleInfo.device], dtype=self._para[ModuleInfo.dtype])
        labels = one_hot.scatter_(0, labels.to(dtype=torch.int64), 1)
        return inputs, labels


class TcvDataset(DatasetTemp):
    """
    torchvision库中的数据集的标准化封装
    """
    def __init__(self, dataset: Any, info: Dict[uuid_t, Any] | None = None) -> None:
        """
        初始化 \n

        :param dataset: tcv数据集，如datasets.CIFAR10 \n
        :param info: `DatasetInfo`，`para` 中无需定义 `train` 参数 \n
        """
        super().__init__(info=info)
        self._container = _TcvDataset
        self._train_dataset = dataset(train=True, **self._info[DatasetInfo.para])
        self._val_dataset = dataset(train=False, **self._info[DatasetInfo.para])




"""
***训练器***
"""


from .const import *
from .trainer import TrainerTemp
from .optimizer import OptimTemp
from .criterion import CriterionTemp
from .evaluator import EvaluatorTemp, EffectorTemp
from .dataloader import DataLoaderTemp
from .dataset import DatasetTemp
from .monitor import TrainerMonitor
from ..temp import ModuleNode



"""
模型容器的异常处理
"""


import warnings


class ModelError(Exception):
    """模型异常"""
    def __init__(self, msg, code=None):
        super().__init__(msg)
        self.code = code


class ModelWarning(Warning):
    """模型警告"""
    pass




"""
**错误或警告**
"""


import warnings


class MEngineCallError(Exception):
    """调用类型未实现"""
    def __init__(self, msg, code=None):
        super().__init__(msg)
        self.code = code


class MEngineCallWarning(Warning):
    pass



"""
模型结构的模板文件 \n
unit - 裸 `nn.Module` 模型 \n
node - 模块节点，最小封装单元，只能包含单一的 `unit` \n
module - 模块，仅包含ModuleKey中的固定术语的单一执行流，可包含固定串行化的 `unit` \n
model - 模型，由若干 `node` 或 `module` 组成，各node构成DAG(有向无环图) \n
"""


from .model import ModelTemp
from .module import ModuleTemp
from .node import ModuleNode
from .unit import ModuleUnit
from .const import ModuleKey, ModuleInfo, ModuleMode



from hagike.models.temp.model import *
from hagike.models.temp.module import *
from hagike.models.temp.node import *
from torchvision.models import *


def test_models_temp_model():
    """models.temp.model的测试用例"""
    weights = EfficientNet_V2_S_Weights.verify(EfficientNet_V2_S_Weights.IMAGENET1K_V1)
    state_dict = weights.get_state_dict()
    model = ModelTemp(
        {
            'all': ModuleNode(efficientnet_v2_s())
        },
        {
            ModelTemp.input_node: ([], ['all']),
            'all': ([ModelTemp.input_node], [ModelTemp.output_node]),
            ModelTemp.output_node: (['all'], [])
         }
    )
    model.load_weights(state_dict, False, key='all')
    model.to(device='cuda', dtype=torch.float32, mode=False)
    model.print_model()
    model.print_summary()


def test_models_temp_module():
    """models.temp.module的测试用例"""
    # 创建模型
    module_dict = {
        ModuleKey.bone: ModuleNode(efficientnet_v2_l())
    }
    model = ModuleTemp(module_dict)
    # 加载权重
    weights = EfficientNet_V2_L_Weights.verify(EfficientNet_V2_L_Weights.IMAGENET1K_V1)
    state_dict = weights.get_state_dict()
    model.load_weights(state_dict, False, key=ModuleKey.bone)
    model.to(device='cuda', dtype=torch.float32, mode=False)
    # 更新结构
    model.update(ModuleKey.bone, ModuleNode(efficientnet_v2_s()))
    weights = EfficientNet_V2_S_Weights.verify(EfficientNet_V2_S_Weights.IMAGENET1K_V1)
    state_dict = weights.get_state_dict()
    model.load_weights(state_dict, False, key=ModuleKey.bone)
    model.to(device='cuda', dtype=torch.float32, mode=False)
    # 打印信息
    model.print_model()
    model.print_summary()


def test_models_temp_node():
    """models.temp.node的测试用例"""
    weights = EfficientNet_V2_S_Weights.verify(EfficientNet_V2_S_Weights.IMAGENET1K_V1)
    state_dict = weights.get_state_dict()
    unit = efficientnet_v2_s()
    module = ModuleNode(unit)
    module.load_weights(state_dict, False)
    module.to(device='cuda', dtype=torch.float32, mode=False)
    module.print_model()
    module.print_summary()



"""
***数据加载器***
"""


from torch.utils.data import DataLoader
from typing import Any, Dict
from .const import uuid_t, DataLoaderInfo
from .dataset import DatasetTemp
from .error import TrainingError


class DataLoaderTemp:
    """数据加载器容器"""

    def __init__(self, dataset: DatasetTemp, info: Dict[uuid_t, Any] | None = None) -> None:
        self._info = DataLoaderInfo.dict_(info)
        self._dataset = dataset
        if self._info[DataLoaderInfo.para]:
            self._info[DataLoaderInfo.train_para] = self._info[DataLoaderInfo.para]
            self._info[DataLoaderInfo.val_para] = self._info[DataLoaderInfo.para]
        self._is_init = False
        self._model_info = None
        self._train = None
        self._val = None

    def init(self, model_info: Dict[uuid_t, Any]):
        """配置目标设备等"""
        if self._is_init:
            return
        self._is_init = True
        self._model_info = model_info
        self._dataset.init(model_info)
        self._train = DataLoader(self._dataset.train, **self._info[DataLoaderInfo.train_para])
        self._val = DataLoader(self._dataset.val, **self._info[DataLoaderInfo.val_para])

    @property
    def train(self) -> DataLoader:
        if self._is_init:
            return self._train
        else:
            raise TrainingError(f"visiting dataloader when DataLoaderTemp uninitialized!!!")

    @property
    def val(self) -> DataLoader:
        if self._is_init:
            return self._val
        else:
            raise TrainingError(f"visiting dataloader when DataLoaderTemp uninitialized!!!")



"""
模型容器中最基本的构成，实现下层库与容器的统一接口 \n
"""


from __future__ import annotations
import torch
import torch.nn as nn
from torchsummary import summary
from typing import Mapping, Any, Sequence
from .error import ModelError, ModelWarning
from hagike.utils import *
from .const import ModuleInfo, uuid_t, ModuleMode


class ModuleNode(nn.Module):
    """
    模型最小组成部分的模板，若希望有自定义实现，可用于被继承 \n
    同时也作为更高级的模型结构的基类 \n
    """

    def __init__(self, model: nn.Module, info: Dict[uuid_t, Any] | None = None):
        """
        初始化 \n

        :param model: 指定模型 \n
        :param info: 指定初始参数描述，若未输入描述则默认为固定值 \n
        """
        super(ModuleNode, self).__init__()
        self._model = model
        self._info = ModuleInfo.dict_(info)
        self.to(device=self._info[ModuleInfo.device],
                dtype=self._info[ModuleInfo.dtype],
                mode=self._info[ModuleInfo.mode])

    @property
    def model(self) -> nn.Module: return self._model
    @property
    def device(self) -> str: return self._info[ModuleInfo.device]
    @property
    def dtype(self) -> torch.dtype: return self._info[ModuleInfo.dtype]
    @property
    def info(self) -> dict: return self._info

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """前向传播"""
        return self._model(x)

    @staticmethod
    def check_node(node: ModuleNode | None) -> None:
        """检查模块是否符合预期，若不是 `ModuleNode` 的子类或 `None` 则报错"""
        if node is None:
            return
        if not isinstance(node, ModuleNode):
            raise ModelError(f"Wrong Module Type, f{type(node)} instead of {ModuleNode}")

    def load_weights(self, weights_src: str | Any, is_path: bool = False) -> None:
        """根据is_path，选择从路径或从内存中加载指定部分的模块参数"""
        if is_path:
            state_dict = torch.load(weights_src, map_location=self.device)
        else:
            state_dict = weights_src
        self._model.load_state_dict(state_dict)

    def save_weights(self, path: str | None = None) -> Any:
        """根据path，选择加载指定部分的模块参数到路径或从内存中"""
        state_dict = self._model.state_dict()
        if path is not None:
            torch.save(state_dict, path)
        return state_dict

    def print_summary(self, input_size=(3, 224, 224)) -> None:
        """
        打印模型的情况，输入尺寸不包括batch，进行模型测试时的参数与当前参数一致 \n
        可用于检查模型可用性 \n
        """
        summary(self, input_size, device=self.device)

    def print_model(self, blank: int = 0) -> None:
        """打印模型构成"""
        blank_str = ' ' * blank
        print(f"{blank_str}ModuleNode: {self._model.__class__.__name__}")

    def to(self,
           device: str | None = None,
           dtype: torch.dtype | None = None,
           mode: uuid_t | None = None) -> None:
        """
        转换模型类型，原位替换 \n
        `mode` 为 `ModuleMode` 型常量，指定模式
        """
        if device is not None:
            self._to_device(device)
        if dtype is not None:
            self._to_dtype(dtype)
        if mode is not None:
            self._to_mode(mode)

    def _to_device(self, device: str) -> None:
        """转换设备"""
        self._info[ModuleInfo.device] = device
        self._model = self._model.to(device=device)

    def _to_dtype(self, dtype: torch.dtype) -> None:
        """转换数据类型"""
        self._info[ModuleInfo.dtype] = dtype
        self._model = self._model.to(dtype=dtype)

    def _to_mode(self, mode: uuid_t) -> None:
        """转换模式：训练 / 评估"""
        ModuleInfo.check_in_(mode)
        self._info[ModuleInfo.mode] = mode
        if mode == ModuleMode.train:
            self._model.train()
        else:
            self._model.eval()



"""
**错误与警告**
"""


import warnings


class MSysInputError(Exception):
    """系统输入类型未实现"""
    def __init__(self, msg, code=None):
        super().__init__(msg)
        self.code = code


class MSysInputWarning(Warning):
    pass


class MSysResponseError(Exception):
    """系统响应错误"""
    def __init__(self, msg, code=None):
        super().__init__(msg)
        self.code = code


class MSysResponseWarning(Warning):
    pass



"""
**`matlab` 引擎管理器** \n
此处说明各种 `matlab` 操作对应的使用方式，假设引擎命名为 `m` \n
直接调用函数或创建实例：\n
result = m.function(arg1, ...) \n
obj = m.class(arg1, ...) \n
调用函数句柄：\n
value = m.feval(methodName, arg1, ...) \n
获取对象成员属性：\n
value = m.subsref(obj, m.substruct('.', propName)) \n
调用类函数：\n
value = m.feval(methodName, obj, arg1, ...) \n
调用特殊字：\n
grid on; \n
m.feval('grid', 'on') \n
"""


import matlab.engine
from hagike.utils.message import add_msg, MsgLevel
from hagike.tools.matlab.scripts import *
from .error import *
from .const import *


class MEngine:
    """`matlab` 引擎"""

    def __init__(self) -> None:
        """检查并创建或链接引擎"""
        if len(matlab.engine.find_matlab()) != 0:
            self._m = matlab.engine.connect_matlab()
        else:
            add_msg(MsgLevel.Warning,
                    f"No Shared Matlab Engine On This Computer. Creating Matlab Engine, It Takes for a While!")
            self._m = matlab.engine.start_matlab()
        self._init_conf()

    def _init_conf(self) -> None:
        """初始化配置，固定配置常用调用接口"""
        self._m.addpath(matlab_script_root)
        self._obj_value = getattr(self._m, 'subsref')
        self._obj_struct = getattr(self._m, 'substruct')
        self._func = getattr(self._m, 'm_feval')
        self._operator = getattr(self._m, 'm_operator')

    def __call__(self, call_type: uuid_t, script: str | uuid_t, *args,
                 num: int = -1, obj: Any = None) -> Any:
        """
        调用引擎 \n
        :param call_type - 调用类型 \n
        :param obj - 对象句柄，若是直接调用类型则此处填 `None` \n
        :param num - 参数数量 \n
        :param script - 函数名称 \n
        :param args - 函数参数，`matlab` 中不支持关键字传参，而只支持顺序传参 \n
        :return - 返回调用结果
        """
        MCall.check_in_(call_type)
        if call_type == MCall.func:
            return self.call(script, *args, num=num)
        elif call_type == MCall.obj_value:
            return self.obj_value(obj, script)
        elif call_type == MCall.obj_func:
            return self.obj_call(script, obj, *args, num=num)
        elif call_type == MCall.print:
            self.print(script, *args)
        elif call_type == MCall.operator:
            return self.operator(script, *args)
        else:
            raise MEngineCallError(f"{MCall.get_name_(call_type)} is not implemented!!!")

    def call(self, script: str, *args, num: int = -1) -> Any:
        """封装直接函数调用"""
        return self._func(script, num, *args)

    def obj_value(self, obj: Any, script: str) -> Any:
        """封装对象属性返回"""
        return self._obj_value(obj, self._obj_struct('.', script))

    def obj_call(self, obj, script: str, *args, num: int = -1) -> Any:
        """封装对象函数调用"""
        return self._func(script, num, obj, *args)

    def print(self, *args) -> None:
        """封装属性打印"""
        self._func('disp', 0, *args)

    def operator(self, script: uuid_t, *arg):
        """封装运算符"""
        base_cls = MOperator.get_base_(script)
        op_type = MOperator.get_name_(base_cls.get_uuid_())
        func = base_cls.get_value_(script)
        return self._operator(op_type, func, *arg)

    @property
    def m(self) -> Any:
        """返回引擎"""
        return self._m

    def exit(self) -> None:
        """断开链接并释放资源"""
        self._m.quit()

    def __del__(self) -> None:
        """释放引擎"""
        self.exit()



"""
系统函数类
"""


from __future__ import annotations
from ..utils import *
from typing import List, Dict, Any
from .const import *
from .error import *
import matplotlib.pyplot as plt
import numpy as np
from ..engine.engine import *


class MSys:
    """`matlab` 系统"""

    def __init__(self, m: MEngine,
                 num: List[float] | None = None, den: List[float] | None = None,
                 z: List[float] | None = None, p: List[float] | None = None, k: float | None = None,
                 fb: MSys | None = None, tf: Any = None) -> None:
        """
        初始化参数，有三种输入方式（直接输入、输入零极点、输入分子分母系数），可选是否进行反馈 \n
        :param m - `matlab` 引擎 \n
        :param tf - 若直接给出 `matlab` 中的 `tf` 系统，则忽略系数项，直接指定
        :param num - 系统函数的分子系数，降幂排列 \n
        :param den - 系统函数的分母系数，降幂排列 \n
        :param z - 零点
        :param p - 极点
        :param k - 增益
        :param fb - 也需要为 `Msys` 类型 \n
        """
        self._m, self._fb = m, fb
        # 获取开环系统
        if tf is not None:
            self._G = tf
            num = np.array(m.obj_value(self._G, 'num'))[0][0]
            den = np.array(m.obj_value(self._G, 'den'))[0][0]
        elif num is not None and den is not None:
            self._G = m.call('tf', m_double(num), m_double(den), num=1)
        elif z is not None and p is not None and k is not None:
            num, den = m.call('zp2tf', m_double(z), m_double(p), m_double(k), num=2)
            num, den = np.array(num)[0], np.array(den)[0]
            self._G = m.call('tf', m_double(num), m_double(den), num=1)
        else:
            raise MSysInputError(f"Incorrect Input Way!!!")

        # 获取闭环系统
        if fb is None:
            self._H = self._G
            self._num, self._den = num, den
        else:
            self._H = m.call('feedback', self._G, fb.H, num=1)
            # 返回类型为 matlab.double 类型，需要将其解包为 numpy 格式
            self._num = np.array(m.obj_value(self._H, 'num'))[0][0]
            self._den = np.array(m.obj_value(self._H, 'den'))[0][0]
        # 获取系统增益
        self._amp = self._num[-1] / self._den[-1] if self._den[-1] != 0.0 else None
        self._t, self._response, self._input_t = None, None, None

    @property   # 返回 `matlab` 系统
    def H(self) -> Any: return self._H
    @property   # 返回系统稳态增益
    def amp(self) -> float | None: return self._amp
    @property   # 返回响应
    def response(self) -> List[float] | None: return self._response

    def _check_response(self, input_t: uuid_t | None = None) -> None:
        """检查是否已经生成指定类型 `response`，若没有则报错"""
        if self._input_t is None:
            raise MSysResponseError(f"No Response Generated Yet!!!")
        if input_t is not None and self._input_t != input_t:
            raise MSysResponseError(
                f"Response {MSysInput.get_name_(self._input_t)} is not "
                f"the asked type {MSysInput.get_name_(input_t)}!!!")

    def gen_response(self, t: List[float], input_t: uuid_t, input_s: List[float] | None = None) -> List[float]:
        """
        生成响应，结果记录在实例中 \n
        :param t - 时刻表 \n
        :param input_t - 输入类型 \n
        :param input_s - 输入系统，当且仅当 `input_t` 为 `else__` 时才有效 \n
        :return 系统响应 \n
        """
        self._t, self._input_t = t, input_t
        MSysInput.check_in_(input_t)
        if input_t == MSysInput.step:
            self._response = self._m.call('step', self._H, m_double(t), num=1)
        elif input_t == MSysInput.impulse:
            self._response = self._m.call('impulse', self._H, m_double(t), num=1)
        elif input_t == MSysInput.ramp:
            self._response = self._m.call('lsim', self._H, m_double(t), m_double(t), num=1)
        elif input_t == MSysInput.else__:
            self._response = self._m.call('lsim', self._H, m_double(input_s), m_double(t), num=1)
        else:
            raise MSysInputError(f"{MSysInput.get_name_(input_t)} is not implemented!!!")
        return self._response

    def plot_response(self, para: Dict[str, Any] | None = None, save_path: str | None = None):
        """绘制响应图，`para` 为绘图配置项"""
        self._check_response()
        if para is None:
            para = dict()
        plt.plot(self._t, self._response, **para)
        plt.grid('on')
        if save_path is not None:
            plt.savefig(save_path)
        plt.show()

    def overshoot(self) -> float:
        """返回超调量"""
        self._check_response(MSysInput.step)
        return self._m.call('overshoot', m_double(self._response), num=1)

    def settlingtime(self, delta: float = 0.05) -> float | None:
        """
        返回稳态响应时间，若响应中未达到稳态条件则返回 `None` \n
        这里假定响应已经达到类稳态状态，使用方法为从后向前扫描响应 \n
        :param delta - 容差 \n
        """
        self._check_response(MSysInput.step)
        if self._amp is None:
            return None
        r_map = np.abs(np.array(self._response) / self.amp - 1) > delta
        r_len = len(r_map)
        first_settle = r_len - np.argmax(~r_map)
        if first_settle == r_len:
            return None
        else:
            return self._t[first_settle]


class MSysConst(MSys):
    """`matlab` 常数块"""
    def __init__(self, m: Any, amp: float = 1.0):
        """创建常数系统"""
        super().__init__(m, [float(amp)], [1.0])




"""
日志消息
"""


from .enum import *
from colorama import Fore, Back, Style


@advanced_enum()
class MsgLevel(SuperEnum):
    Run = Fore.GREEN + "RUN: "
    Warning = Fore.YELLOW + "WARNING: "
    Error = Fore.RED + "ERROR: "
    Panic = Fore.RED + "PANIC: "


def add_msg(level: int, script: str, is_print=True):
    """添加消息"""
    if is_print:
        msg = script
        MsgLevel.check_in_(level)
        msg = MsgLevel.get_value_(level) + msg + Style.RESET_ALL
        print(msg)


def error_proc(is_exit=True):
    """错误处理"""
    if is_exit:
        exit(-1)




"""
噪声添加
"""

import cv2
import numpy as np
from hagike.utils import *
from .file.image import *
from typing import Mapping, Any


class ImageNoiseError(Exception):
    """图像噪声异常"""
    def __init__(self, msg, code=None):
        super().__init__(msg)
        self.code = code


@advanced_enum()
class ImNoise(SuperEnum):
    """
    噪声类型 \n
    .. important::
        参数是在图像归一化情况下的
    """

    class gaussian(SuperEnum):
        """高斯噪声"""
        mean: uuid_t = 0
        """均值"""
        var: uuid_t = 0.01
        """方差"""

    class salt_pepper(SuperEnum):
        """椒盐噪声"""
        prob_salt: uuid_t = 0.02
        """白色块概率"""
        prob_pepper: uuid_t = 0.02
        """黑色块概率"""

    class poisson(SuperEnum):
        """泊松噪声"""
        scale: uuid_t = 0.12
        """
        泊松分布是一种离散概率分布，用于预测在固定时间或空间内发生某事件的平均次数。 \n
        在图像处理中，泊松噪声通常用于模拟光子到达传感器的随机性，特别是在低光照条件下。 \n
        将图像的每个像素值除以一个尺度参数 `scale`，调整图像数据的“亮度”，以便与泊松分布的参数相匹配。 \n
        `scale` 参数在这里起到了调节噪声强度的作用，较小的 `scale` 值会导致更强烈的噪声 \n
        当 `scale` 足够大时，趋向于椒盐噪声中白点的效果果，且随着 `scale` 继续增大，噪声效果变小 \n
        因为每个像素值被“缩放”得更小，然后泊松分布会在这个更小的值上产生更大的随机变化。 \n
        最后将上一步生成的泊松随机数乘以scale，缩放回原尺度。 \n
        """
        alpha: uuid_t = 0.3
        """
        附加比例因子，用于控制噪声相比原图的占比
        """


def add_noise_to_image(im: ImStd, noise: uuid_t, para: Mapping[uuid_t, Any] | None = None) -> ImStd:
    """
    向图像添加指定类型的噪声
    """
    ImNoise.check_in_(noise)
    noise_type = ImNoise.get_cls_(noise)
    para = noise_type.dict_(para, is_force=True)
    # 保存旧值，这里需要指定 `is_new=True` 来确保操作过程对输入项没有影响
    scale = im.scale
    im = im.to_scale(ImScale.float_n, is_new=True)
    image = im.image
    # 将图像转换到[0, 1]之间
    if noise == ImNoise.gaussian.get_uuid_():
        sigma = para[ImNoise.gaussian.var] ** 0.5
        gaussian = np.random.normal(para[ImNoise.gaussian.mean], sigma, image.shape).astype(np.float32)
        noisy = np.clip(image + gaussian, 0.0, 1.0)
    elif noise == ImNoise.salt_pepper.get_uuid_():
        rand_mask = np.random.rand(*image.shape)
        salt = rand_mask < para[ImNoise.salt_pepper.prob_salt]
        pepper = rand_mask > 1 - para[ImNoise.salt_pepper.prob_pepper]
        image[salt] = 1.0
        image[pepper] = 0.0
        noisy = image
    elif noise == ImNoise.poisson.get_uuid_():
        lam = para[ImNoise.poisson.scale]
        poisson = np.random.poisson(image / lam).astype(np.float32) * lam * para[ImNoise.poisson.alpha]
        noisy = np.clip(image + poisson, 0.0, 1.0)
    else:
        raise ImageNoiseError(
            f"Noise {ImStyle.get_name_(noise)} is not implemented a noise function!!!")
    # 还原为输入时的数值范围
    im.image = noisy
    im.to_scale_(scale)
    return im



"""
***优化器容器***
"""


import torch.optim as optim
from .const import uuid_t, OptimInfo
from ..temp import ModuleNode
from .error import TrainingError
from typing import Dict, Any


class OptimTemp:
    """优化器容器"""

    def __init__(self, info: Dict[uuid_t, Any] | None = None) -> None:
        """创建优化器容器，容器需要先外部初始化才能使用"""
        self._info = OptimInfo.dict_(info)
        self._op = None
        self._is_init = False

    def init(self, model: ModuleNode) -> None:
        """初始化优化器"""
        if self._is_init:
            return
        self._is_init = True
        self._op = self._info[OptimInfo.op_type](
            model.parameters(), lr=self._info[OptimInfo.lr],
            **self._info[OptimInfo.para]
        )

    @property
    def op(self) -> optim.Optimizer:
        """返回优化器"""
        if self._is_init:
            return self._op
        else:
            raise TrainingError(f"visiting op when optimizer uninitialized!!!")

    def __getattr__(self, name: str):
        """重定向属性或方法至 `self._op`"""
        return getattr(self.op, name)



"""
***图像滤波*** \n
"""

import cv2
import numpy as np
import matplotlib.pyplot as plt
from .file.image import *
from hagike.utils import *


class ImageFilterError(Exception):
    """图像滤波器异常"""

    def __init__(self, msg, code=None):
        super().__init__(msg)
        self.code = code


@advanced_enum()
class ImSmooth(SuperEnum):
    """平滑滤波"""

    class mean(SuperEnum):
        """均值滤波"""
        ksize: uuid_t = (5, 5)

    class gaussian(SuperEnum):
        """高斯滤波"""
        sigma: uuid_t = 1


def apply_smoothing_filter(im: ImStd, smooth: uuid_t, para: Mapping[uuid_t, Any] | None = None) -> ImStd:
    """对图像应用平滑滤波"""
    ImSmooth.check_in_(smooth)
    smooth_type = ImSmooth.get_cls_(smooth)
    para = smooth_type.dict_(para, is_force=True)
    style, scale = im.style, im.scale
    im = im.to_style(ImStyle.cv_ndarray, is_new=True)
    im.to_scale_(ImScale.float_n)
    image = im.image

    if smooth == ImSmooth.mean.get_uuid_():
        image = cv2.blur(image, para[ImSmooth.mean.ksize])
    elif smooth == ImSmooth.gaussian.get_uuid_():
        image = cv2.GaussianBlur(image, (0, 0), para[ImSmooth.gaussian.sigma])
    else:
        raise ImageFilterError(
            f"Filter {ImSmooth.get_name_(smooth)} is not implemented a filter function!!!")

    im.image = np.clip(image, 0.0, 1.0)
    im.to_style_(style), im.to_scale_(scale)
    return im


@advanced_enum()
class ImSharp(SuperEnum):
    """锐化滤波"""

    class laplacian(SuperEnum):
        """拉普拉斯滤波"""
        ksize: uuid_t = 5

    class sharpening(SuperEnum):
        """锐化滤波"""
        alpha: uuid_t = 1.5
        beta: uuid_t = -0.5


def apply_sharpening_filter(im: ImStd, sharp: uuid_t, para: Mapping[uuid_t, Any] | None = None) -> ImStd:
    """对图像应用锐化滤波"""
    ImSharp.check_in_(sharp)
    sharp_type = ImSharp.get_cls_(sharp)
    para = sharp_type.dict_(para, is_force=True)
    style, scale = im.style, im.scale
    im = im.to_style(ImStyle.cv_ndarray, is_new=True)
    im.to_scale_(ImScale.float_n)
    image = im.image

    if sharp == ImSharp.laplacian.get_uuid_():
        laplacian = cv2.Laplacian(image, cv2.CV_32F, ksize=para[ImSharp.laplacian.ksize])
        image = image - laplacian
    elif sharp == ImSharp.sharpening.get_uuid_():
        image = para[ImSharp.sharpening.alpha] * image + para[ImSharp.sharpening.beta]
    else:
        raise ImageFilterError(
            f"Filter {ImSharp.get_name_(sharp)} is not implemented a filter function!!!")

    im.image = np.clip(image, 0.0, 1.0)
    im.to_style_(style), im.to_scale_(scale)
    return im


@advanced_enum()
class ImFreq(SuperEnum):
    """频域滤波"""

    class LPF(SuperEnum):
        """理想低通"""
        radius: uuid_t = 0.5
        """半径相比原图半径的比例"""

    class HPF(SuperEnum):
        """理想高通"""
        radius: uuid_t = 0.5
        """半径相比原图半径的比例"""


def frequency_domain_filtering(im: ImStd, freq: uuid_t, para: Mapping[uuid_t, Any] | None = None) -> ImStd:
    """频域滤波"""
    ImFreq.check_in_(freq)
    freq_type = ImFreq.get_cls_(freq)
    para = freq_type.dict_(para, is_force=True)
    style, scale, color = im.style, im.scale, im.color
    im = im.to_style(ImStyle.im_ndarray, is_new=True)
    im.to_scale_(ImScale.float_n), im.to_color_(ImColor.freq)
    image = im.image

    if freq == ImFreq.LPF.get_uuid_():
        rows, cols = image.shape
        c_row, c_col = rows // 2, cols // 2
        r_radius, c_radius = int(para[ImFreq.LPF.radius] * rows), int(para[ImFreq.LPF.radius] * cols)
        mask = np.zeros((rows, cols), np.uint8)
        cv2.ellipse(mask, (c_col, c_row), (c_radius, r_radius), 
                    angle=0, startAngle=0, endAngle=360, color=(1, 1, 1), thickness=-1)
        image = image * mask
    elif freq == ImFreq.HPF.get_uuid_():
        rows, cols = image.shape
        c_row, c_col = rows // 2, cols // 2
        r_radius, c_radius = int(para[ImFreq.HPF.radius] * rows), int(para[ImFreq.HPF.radius] * cols)
        mask = np.ones((rows, cols), np.uint8)
        cv2.ellipse(mask, (c_col, c_row), (c_radius, r_radius),
                    angle=0, startAngle=0, endAngle=360, color=(0, 0, 0), thickness=-1)
        image = image * mask
    else:
        raise ImageFilterError(
            f"Filter {ImFreq.get_name_(freq)} is not implemented a filter function!!!")

    im.image = image
    im.to_style_(style), im.to_scale_(scale), im.to_color_(color)
    return im



from hagike.utils.enum import *


def test_utils_enum():
    """utils.enum的测试用例"""

    # 举例
    @advanced_enum()
    class EnumExample(SuperEnum):
        """一个使用创建枚举类型的例子"""
        _sequence = (
            'z', 'b', 'SubExample1'
        )
        a__ = 0
        z = 2
        b = 3

        class SubExample1(SuperEnum):
            _value = 4
            a = 5
            b = 6

        class SubExample2(SuperEnum):
            _sequence = (
                'c', 'd', 'SubSubExample'
            )
            _hide = True
            c = 7
            d = 8

            class SubSubExample(SuperEnum):
                e = 9
                f = 10

    # 测试
    EnumExample.SubExample2.print_(is_value=True)
    EnumExample.tree_(is_value=True)
    print(EnumExample.dict_())
    for i in EnumExample.iter_():
        print(i, end=', ')
    print()
    print(EnumExample.list_({EnumExample.b: 5, EnumExample.z: 6}))




"""
***与文件处理相关的库函数*** \n
"""


import os


class FileWritableError(Exception):
    """文件不可写异常"""
    def __init__(self, msg, code=None):
        super().__init__(msg)
        self.code = code


class FileReadableError(Exception):
    """文件不可读异常"""
    def __init__(self, msg, code=None):
        super().__init__(msg)
        self.code = code


def check_path_readable(path: str, is_raise: bool = True) -> bool:
    """
    确保目标可读取，检查项包括： \n
    0. 检查路径本身的合法性 \n
    1. 路径本身存在 \n
    2. 路径是文件而非文件夹 \n
    3. 路径有读取权限 \n
    """
    # 0
    try:
        os.path.dirname(path)
    except TypeError:
        if is_raise:
            raise FileReadableError(f"ERROR: {path} is not a string!!!")
        else:
            return False
    # 1
    if not os.path.exists(path):
        if is_raise:
            raise FileReadableError(f"ERROR: {path} does not exist!!!")
        else:
            return False
    # 2
    if os.path.isdir(path):
        if is_raise:
            raise FileReadableError(f"ERROR: {path} is a dir!!!")
        else:
            return False
    # 3
    if os.access(path, os.R_OK):
        return True
    else:
        if is_raise:
            raise FileReadableError(f"ERROR: {path} is not Readable, Permission Denied!!!")
        else:
            return False


def ensure_path_writable(path: str, is_raise: bool = True) -> bool:
    """
    确保目标路径可写入，包括检查项： \n
    0. 检查路径本身的合法性 \n
    1. 路径本身不是一个已存在的文件夹 \n
    2. 确保路径的父文件夹存在，若不存在则创建 \n
    3. 确保其有写入权限 \n
    如果指定 `is_raise=True`，则会在不可达时触发异常，否则仅返回 `False`
    """
    # 0
    try:
        directory = os.path.dirname(path)
    except TypeError:
        if is_raise:
            raise FileWritableError(f"ERROR: {path} is not a string!!!")
        else:
            return False
    # 1
    if os.path.isdir(path):
        if is_raise:
            raise FileWritableError(f"ERROR: {path} is a dir!!!")
        else:
            return False
    # 2
    try:
        os.makedirs(directory, exist_ok=True)
    except PermissionError:
        if is_raise:
            raise FileWritableError(f"ERROR: dir {directory} is cannot be created, Permission Denied!!!")
        else:
            return False
    # 3
    if os.access(directory, os.W_OK):
        return True
    else:
        if is_raise:
            raise FileWritableError(f"ERROR: dir {directory} is not Writable, Permission Denied!!!")
        else:
            return False







"""
***监视器***
"""


from datetime import datetime
from torch.utils.tensorboard import SummaryWriter
from .const import uuid_t, MonitorInfo, TrainerInfo
from typing import Dict, Any
from hagike.log import LoggerTemp, LogTemp, LoggerConf
import os
from tqdm import tqdm
from .error import TrainingError


class TrainerMonitor:
    """
    ***模型监视器***
    """

    def __init__(self, info: Dict[uuid_t, Any] | None = None) -> None:
        """
        按照参数配置监视器，默认监视器实现如下功能： \n
        1. 提供各种可能用到的统一接口，提供最基本的接口规范 \n
        2. 提供一个最基本的实现，通过日志方式提供记录与分析功能，控制终端输出与显示 \n
        3. 接入外部监控器，如 `Tensorboard` \n
        注意：在创建监视器后必须在外部进行init才能正常工作！！！ \n

        :param info: `MonitorInfo` 监视器参数 \n
        """
        self._info = MonitorInfo.dict_(info)
        # 创建日志子目录
        now_date = datetime.now()
        formatted_date = now_date.strftime("%Y-%m-%d-%H-%M-%S")
        self._info[MonitorInfo.logdir] = os.path.join(self._info[MonitorInfo.logdir], formatted_date)
        # todo: 根据TrainerMonitor的需求配置LoggerTemp
        self._logger = LoggerTemp(is_init=False, conf={
            LoggerConf.logdir: self._info[MonitorInfo.logdir]
        })
        self._logger.init()
        # 这里默认使用 `Tensorboard` 作为外部监视器
        os.makedirs(self._info[MonitorInfo.logdir], exist_ok=True)
        self._writer = SummaryWriter(self._info[MonitorInfo.logdir])
        self._is_init = False
        self._train_info: Dict[uuid_t, Any] | None = None
        self._pbar: tqdm | None = None

    def init(self, train_info: Dict[uuid_t, Any]) -> None:
        """接入训练参数，初始化监视器"""
        if self._is_init:
            return
        self._is_init = True
        self._train_info = train_info
        # 用子进程打开
        if self._info[MonitorInfo.autostart]:
            import subprocess
            process = subprocess.Popen(
                [f'konsole --hold -e tensorboard --logdir={self._info[MonitorInfo.logdir]}'],
                shell=True, close_fds=True)
        # todo: 在日志中保存训练参数，并在日志文件夹中对训练参数进行备份

    @property
    def writer(self) -> SummaryWriter:
        if self._is_init:
            return self._writer
        else:
            raise TrainingError(f"visiting writer when train monitor uninitialized!!!")

    @property
    def logger(self) -> LoggerTemp:
        if self._is_init:
            return self._logger
        else:
            raise TrainingError(f"visiting logger when train monitor uninitialized!!!")

    @property
    def pbar(self) -> tqdm:
        if self._is_init:
            return self._pbar
        else:
            raise TrainingError(f"visiting pbar when trainer not started!!!")

    @property
    def logdir(self) -> str:
        return self._info[MonitorInfo.logdir]

    def start(self):
        """开始训练"""
        self._logger.add_log(LogTemp(msg="Begin Training Process"))
        self._pbar = tqdm(total=self._train_info[TrainerInfo.max_epochs])

    def new_epoch(self, epoch: int):
        """记录新一轮训练"""
        self.pbar.update()
        self.logger.add_log(LogTemp(msg=f"Begin Training Epoch {epoch}"), is_print=False)

    def record_effect(self, epoch: int, train_loss: float, val_loss: float, effect: float):
        """记录训练效果"""
        self.logger.add_log(
            LogTemp(msg=f"Epoch {epoch}: train - {train_loss}; val - {val_loss}; effect - {effect}"), is_print=False)
        # todo: 通过配置方法定义名称
        self.writer.add_scalar('Train Loss', train_loss, epoch)
        self.writer.add_scalar('Val Loss', val_loss, epoch)
        self.writer.add_scalar('Effect', effect, epoch)

    def end(self):
        """关闭监视器"""
        self.pbar.close()
        self.logger.add_log(LogTemp(msg="End Training Process"))

    def destroy(self):
        """销毁监视器"""
        self.writer.close()
        self.logger.end()



"""
模型容器的常量表
"""

from hagike.utils.enum import uuid_t, SuperEnum, advanced_enum


@advanced_enum()
class TrainerInfo(SuperEnum):
    """训练参数表"""
    max_epochs = 100
    """最大轮数"""
    device = 'cuda'
    """训练设备"""
    others = dict()
    """其它"""


@advanced_enum()
class EvaluatorInfo(SuperEnum):
    """评估器参数表"""
    effector_type = None
    """效果生成器类型"""
    para = dict()
    """Dict[str, Any], 效果参数"""
    others = dict()
    """其它"""


@advanced_enum()
class MonitorInfo(SuperEnum):
    """监视器参数表"""
    logdir = None
    """监视器日志的存储位置"""
    autostart = True
    """是否自动启动外置监视进程"""
    others = dict()
    """其它"""


@advanced_enum()
class DatasetInfo(SuperEnum):
    """数据集参数表"""
    train_rate = None
    """训练集数量占比"""
    para = dict()
    """Dict[str, Any], Dataset类一次封装的配置参数（外部创建参数）"""
    sub_para = dict()
    """Dict[str | uuid_t, Any], Dataset类二次封装的配置参数（内部container配置参数）"""
    others = dict()
    """其它"""


@advanced_enum()
class DataLoaderInfo(SuperEnum):
    """数据加载器参数表"""
    para = dict()
    """Dict[str, Any], 如果定义了para，则默认训练部分与评估部分参数配置一致"""
    train_para = dict()
    """训练部分DataLoader的参数表"""
    val_para = dict()
    """评估部分DataLoader的参数表"""
    others = dict()
    """其它"""


@advanced_enum()
class OptimInfo(SuperEnum):
    """训练参数表"""
    lr = 0.001
    """学习率"""
    op_type = None
    """优化器类型"""
    para = dict()
    """Dict[str, Any], 优化器参数"""
    others = dict()
    """其它"""


@advanced_enum()
class CriterionInfo(SuperEnum):
    """损失函数参数表"""
    crt_type = None
    """损失函数类型"""
    para = dict()
    """Dict[str, Any], 损失函数参数"""
    others = dict()
    """其它"""



"""
***评估器***
"""


import torch
import torch.nn as nn
from .const import uuid_t, EvaluatorInfo
from typing import Dict, Any, Tuple
from ..temp import ModuleNode, ModuleMode
from .error import TrainingError
from .dataloader import DataLoaderTemp
from .criterion import CriterionTemp


class EffectorTemp:
    """效果生成器，这里的基准类实现了分类网络的准确率评估"""
    def __init__(self):
        """初始化"""
        self._activator = nn.Softmax(dim=-1)

    def set(self):
        """生成初始 `effect`"""
        return 0.0

    def gen(self, outputs: torch.Tensor, labels: torch.Tensor):
        """
        单次获取 `effect` \n
        :param outputs: (batch_size, num_classes), 独热码, 0.0 ~ 1.0, float32
        :param labels: (batch_size, num_classes), 独热码, 1.0, float32
        """
        # torch..max 返回 (值, 索引)
        _, predicted = torch.max(self._activator(outputs), 1)
        _, labeled = torch.max(labels, 1)
        return (predicted == labeled).sum()

    def add(self, a, b):
        """效果累积函数"""
        return a + b

    def avg(self, effect, num):
        """平均化效果"""
        return effect / num


class EvaluatorTemp:
    """评估器"""

    def __init__(self, info: Dict[uuid_t, Any] | None = None):
        """创建评估器容器"""
        self._info = EvaluatorInfo.dict_(info)
        self._effector: EffectorTemp = self._info[EvaluatorInfo.effector_type](**self._info[EvaluatorInfo.para])
        self._model = None
        self._dataloader = None
        self._criterion = None
        self._is_init = False

    def init(self, model: ModuleNode, dataloader: DataLoaderTemp, criterion: CriterionTemp):
        """初始化，配置模型与测试数据集"""
        if self._is_init:
            return
        self._is_init = True
        self._model, self._dataloader, self._criterion = model, dataloader, criterion

    def evaluate(self) -> Tuple[float, Any]:
        """评估过程封装"""
        if self._is_init:
            return self._evaluate()
        else:
            raise TrainingError(f"Evaluate when container uninitialized!!!")

    def _evaluate(self) -> Tuple[float, Any]:
        self._model.to(mode=ModuleMode.val)
        val_loss, sample_num = 0.0, 0
        val_effect = self._effector.set()
        with torch.no_grad():
            for i, data in enumerate(self._dataloader.val):
                inputs, labels = data
                outputs = self._model(inputs)
                loss = self._criterion(outputs, labels)
                effect = self._effector.gen(outputs, labels)
                val_loss += loss * inputs.size(0)
                sample_num += inputs.size(0)
                val_effect = self._effector.add(val_effect, effect)
            val_loss /= sample_num
            val_effect = self._effector.avg(val_effect, sample_num)
        return val_loss, val_effect



"""
***模型的父亲类模板*** \n
"""


import torch.nn as nn
from typing import Any, Dict, Tuple, List, Mapping, Sequence
from .const import uuid_t
from .error import ModelError, ModelWarning
from .node import ModuleNode
from copy import deepcopy


# 节点标识符，设置为 `str` 是因为 `nn.ModuleDict` 的键值要求为 `str`
node_t = str


class ModelTemp(ModuleNode):
    """模型的通用模板父类"""

    input_node = '_in'
    output_node = '_out'
    _in = 0
    _out = 1
    _blank = 4

    def __init__(self, nodes: Mapping[node_t, ModuleNode],
                 deps: Dict[node_t,  Tuple[List[node_t], List[node_t]]],
                 info: Dict[uuid_t, Any] | None = None) -> None:
        """
        根据输入构建有向无环图方式的模型 \n

        :param nodes: 节点映射表，键值不能与 `input_node` 与 `output_node` 的键值重复 \n
        :param deps: 节点依赖关系表，每个节点都会描述自身的 `(输入节点, 输出节点)`；
                    要求必须含有 `input_node` 与 `output_node` 的键值，
                    前者的 `输入节点` 要求为 `[]`，后者的 `输出节点` 为 `[]` \n
        :param info: 指定初始参数描述，若未输入描述则默认为固定值 \n
        """
        self._deps = deps
        keys, values = nodes.keys(), nodes.values()
        for node in values:
            self.check_node(node)
        if (self.input_node in keys) or (self.output_node in keys):
            raise ModelError(f"{self.input_node} or {self.output_node} should not be used as a key!!!")
        keys = deps.keys()
        if (self.input_node not in keys) or (self.output_node not in keys):
            raise ModelError(f"{self.input_node} or {self.output_node} should be included in deps!!!")
        super().__init__(nn.ModuleDict(nodes), info)
        self._check_model()
        # 临时依赖表，存放依赖关系未满足的节点
        self._dep_dict: Dict[node_t, Tuple[List[node_t] | None, List[node_t] | None]] | None = None
        # 临时等待队列，存放可以进行运算的节点
        self._wait_list: List[node_t] | None = None
        # 数据暂存队列，存储节点的输入数据
        self._store_dict: Dict[node_t, List[Any]] | None = None
        self._store_default = dict()
        for node in self._deps.keys():
            self._store_default[node] = [None for _ in range(len(self._deps[node][self._in]))]

    def _check_model(self) -> None:
        """
        .. todo::
            检查模型合法性，包括其是否成环，以及输入输出数量是否匹配
        """

    def update(self) -> None:
        """
        .. todo::
            更新模型结构
        """

    def _reset_tmp(self) -> None:
        """重置临时结构"""
        self._dep_dict = deepcopy(self._deps)
        self._wait_list = list()
        self._store_dict = deepcopy(self._store_default)

    def _release_tmp(self) -> None:
        """释放临时结构"""
        self._dep_dict = None
        self._wait_list = None
        self._store_dict = None

    def print_model(self, blank: int = 0) -> None:
        """打印模型构成"""
        blank_str = ' ' * blank
        print(f"{blank_str}ModelTemp: {self.__class__.__name__}")
        blank += self._blank
        blank_str += ' ' * self._blank
        for key, node in self._model.items():
            print(f"{blank_str}{key}：")
            node.print_model(self._blank + blank)

    def _step_node(self, this_node: node_t, out_node: List[node_t], results: Sequence[Any]) -> None:
        """处理节点"""
        for i, node in enumerate(out_node):
            # 获取输出位置的输入节点，据此载入至数据暂存区的正确位置
            in_node = self._dep_dict[node][self._in]
            index = in_node.index(this_node)
            in_node.pop(index)
            self._store_dict[node][index] = results[i]
            # 检查该节点是否已经满足依赖条件，若满足则加入等待队列
            if not in_node:
                self._wait_list.append(node)

    def forward(self, *args) -> Tuple | Any:
        """
        :param args: 输入元组，会依次展开给第一级节点
        :return: 输出列表
        """
        # 重置临时结构
        self._reset_tmp()
        outputs = None
        # 解析并遍历入口节点的输出位置
        out_node = self._dep_dict[self.input_node][self._out]
        results = args
        # 遍历输出位置
        self._step_node(self.input_node, out_node, results)
        # 前向传播过程，等待表非空则持续进行下去
        while self._wait_list:
            # 从等待表获取待处理节点，获取后入先出，优先执行最近路径
            this_node = self._wait_list.pop()
            # 检查节点是否为输出节点，如果是则退出循环
            if this_node == self.output_node:
                outputs = self._store_dict[this_node]
                break
            # 处理节点并获取结果
            result = self._model[this_node](*self._store_dict.pop(this_node))
            # 解析当前节点的输出位置
            out_node = self._dep_dict[this_node][self._out]
            results = [result for _ in range(len(out_node))]
            # 遍历输出位置
            self._step_node(this_node, out_node, results)
        # 整理输出结果
        if outputs is None:
            raise ModelError(f"the model output None, something went wrong!!!")
        elif len(outputs) == 1:
            outputs = outputs[0]
        else:
            outputs = tuple(outputs)
        # 释放临时结构
        self._release_tmp()
        return outputs

    def module(self, key: node_t) -> ModuleNode:
        """返回模块"""
        return self._model[key]

    def load_weights(self, weights_src: str | Any, is_path: bool = False, key: node_t = 'all__') -> None:
        """
        根据is_path，选择从路径或从内存中加载指定部分的模块参数 \n
        `key` 指定模块标识符，若为 `all__` 表示整个模型
        """
        if key == 'all__':
            load_weights = super().load_weights
        else:
            load_weights = self._model[key].load_weights
        return load_weights(weights_src, is_path)

    def save_weights(self, path: str | None = None, key: node_t = 'all__') -> Any:
        """
        根据path，选择加载指定部分的模块参数到路径或从内存中 \n
        `key` 指定模块标识符，若为 `all__` 表示整个模型
        """
        if key == 'all__':
            save_weights = super().save_weights
        else:
            save_weights = self._model[key].save_weights
        return save_weights(path)








"""
常量表
"""


from hagike.utils.enum import SuperEnum, advanced_enum, uuid_t


@advanced_enum()
class MSysInput(SuperEnum):
    """输入类型"""
    impulse = None
    """冲激响应"""
    step = None
    """阶跃响应"""
    ramp = None
    """斜坡响应"""
    else__ = None
    """非标准类型"""



"""
***日志常量表***
"""


from hagike.utils.enum import uuid_t, advanced_enum, SuperEnum


@advanced_enum()
class LoggerConf(SuperEnum):
    """日志管理器配置表"""
    logdir = None
    """日志输入文件夹，若配置为 `None` 则不进行文件输出"""
    event_list = dict()
    """提供事件号到事件列表的映射"""
    others = dict()
    """其它项"""


@advanced_enum()
class LoggerStatus(SuperEnum):
    """日志管理器状态表"""
    inited = None
    """已创建"""
    running = None
    """正在运行"""
    stopped = None
    """已暂停"""
    ended = None
    """已结束"""
    destroyed = None
    """已销毁"""



"""
**常量表**
"""

from hagike.utils.enum import *


@advanced_enum()
class MCall(SuperEnum):
    """调用引擎的类型"""
    func = None
    """调用函数或创建实例或函数化封装的特殊关键字或调用函数句柄"""
    obj_value = None
    """对象成员属性"""
    obj_func = None
    """对象成员函数"""
    print = None
    """打印属性"""
    operator = None
    """运算符"""


# 将算术运算符转为函数，包括：
# 基本算术运算符（Basic）
#    加法（Addition）：+
#    减法（Subtraction）：-
#    乘法（Multiplication）：*
#    除法（Division）：/
#    左除（Left Division）：\（左除，用于矩阵 Ax = B，x = B \ A）
#    幂运算（Power）：^
# 点运算符（Dot）
#    点乘（Element-wise Multiplication）：.*
#    点除（Element-wise Division）：./
#    点幂（Element-wise Power）：.^
# 矩阵运算符（Matrix）
#    共轭转置（Conjugate Transpose）：'
#    非共轭转置（Non-conjugate Transpose）：.'
# 逻辑运算符（Logic）
#    与（AND）：&
#    或（OR）：|
#    非（NOT）：~
#    短路与（SAND）：&&
#    短路或（SOR）：||
#    大于（Bigger）：>
#    大于等于（Bigger Equal）：>=
#    小于（Smaller）: <
#    小于等于（Smaller Equal）: <=
#    等于（Equal）：==
#    不等于（Unequal）：~=
@advanced_enum()
class MOperator(SuperEnum):
    """操作符的类型，类别名字与 `m_operator.m` 中一致"""

    class basic(SuperEnum):
        add = '+'
        sub = '-'
        mul = '*'
        div = '/'
        ldiv = '\\'
        pow = '^'

    class dot(SuperEnum):
        elmul = '.*'
        eldiv = './'
        elpow = '.^'

    class matrix(SuperEnum):
        conj_transpose = "'"
        nonconj_transpose = ".'"

    class logic(SuperEnum):
        and_m = '&'
        or_m = '|'
        not_m = '~'
        and_short = '&&'
        or_short = '||'
        gt = '>'
        ge = '>='
        lt = '<'
        le = '<='
        eq = '=='
        ne = '~='



"""
图像变换
"""


from .file.image.im import *
from PIL import Image
from typing import Callable
import math


def transform_curve(im: ImStd, transform: Callable) -> ImStd:
    """对图像强度进行整体变换，仅可用于 `im_file`，本地转换"""
    im.image = Image.eval(im.image, transform)
    return im


def im_log(im: ImStd) -> ImStd:
    """对数变换，仅可用于 uint8，np，本地转换"""
    scale = 255 / math.log1p(255)
    im.image = np.log1p(im.image) * 255 / math.log1p(255)
    return im



"""
***数据包封装*** \n
.. todo::
    完善压缩类型 \n
"""

from .const import *
from .error import *
from typing import Mapping, Any, Sequence
from hagike.utils.file import *
from hagike.utils.cache import *


class DataPackTemp:
    """数据包模板"""

    def __init__(self, data: Any, src: uuid_t = DataSrc.mem, form: uuid_t = DataForm.direct,
                 info: Dict[uuid_t, Any] = None) -> None:
        """
        创建数据包

        :param data 数据描述符，根据 `src` 进行解析 \n
        :param src 数据源类型
        :param form 数据格式
        :param info 数据额外信息
        """
        DataSrc.check_in_(src), DataForm.check_in_(form)
        info = DataInfo.dict_(info)
        self._data, self._src, self._form, self._info = data, src, form, info

    @property
    def data(self) -> Any:
        self.use()
        return self._data

    @data.setter
    def data(self, data: Any) -> None: self._data = data
    @property
    def is_mem(self) -> bool: return self._src == DataSrc.mem
    @property
    def is_direct(self) -> bool: return self._form == DataForm.direct
    @property
    def is_usable(self) -> bool: return self.is_direct and self.is_mem

    def check(self) -> None:
        """确认数据可达性，若不可达则报错，如果 `_src` 类型为 `else__`，则要求在 `_info` 中自定义检查函数"""
        if self._src == DataSrc.mem:
            return
        elif self._src == DataSrc.file:
            check_path_readable(self._data)
        elif self._src == DataSrc.else__:
            func = self._info[DataInfo.load_check]
            if func is not None:
                func(self._data)
            else:
                raise DataPackError(f"Check func for Data src type 'else__' is not included in info")
        else:
            raise DataPackError(f"Check func for src type {DataSrc.get_name_(self._src)} is not implemented")

    def use(self) -> None:
        """使数据直接可用"""
        self.load()
        self.transfer()

    def load(self) -> None:
        """加载数据包至内存"""
        self.check()
        if self._src == DataSrc.mem:
            return
        elif self._src == DataSrc.file:
            self._data = load_data_from_pkl(self._data)
        elif self._src == DataSrc.else__:
            func = self._info[DataInfo.load_func]
            if func is not None:
                self._data = func(self._data)
            else:
                raise DataPackError(f"Load func for Data src type 'else__' is not included in info")
        else:
            # todo: 压缩类型的加载
            raise DataPackError(f"Load func for src type {DataSrc.get_name_(self._src)} is not implemented")
        self._src = DataSrc.mem

    def transfer(self) -> None:
        """转换至直接可用格式"""
        if self._form == DataForm.direct:
            return
        elif self._form == DataForm.mem_cached:
            func = self._info[DataInfo.mem_cacher]
            if func is not None:
                self._data = func(self._data)
            else:
                raise DataPackError(f"Transfer func for Data form 'mem_cached' is not included in info")
        elif self._form == DataForm.else__:
            func = self._info[DataInfo.form_func]
            if func is not None:
                self._data = func(self._data)
            else:
                raise DataPackError(f"Transfer func for Data form 'else__' is not included in info")
        else:
            raise DataPackError(f"Check func for form {DataSrc.get_name_(self._src)} is not implemented")
        self._form = DataForm.direct

    def to(self) -> None:
        """
        .. todo::
            转换数据源或格式
        """





"""
配置类
"""


import os
import json
from .message import *
from typing import Any


def read_json(json_path: str) -> dict:
    """读取json文件并转换为字典"""
    data = dict()
    try:
        with open(json_path, 'r') as config_file:
            try:
                data = json.load(config_file)
            except json.JSONDecodeError:
                add_msg(MsgLevel.Error.value, f"read_json - File '{json_path}' Not Json Code")
                error_proc()
    except FileNotFoundError:
        add_msg(MsgLevel.Error.value, f"read_json - File '{json_path}' Not Found")
        error_proc()
    return data


def write_json(json_path: str, data: dict, settings: dict | None = None) -> None:
    """写入json文件并转换为字典"""
    dir_name = os.path.dirname(json_path)
    os.makedirs(dir_name, exist_ok=True)
    with open(json_path, 'w', encoding='utf-8') as json_file:
        json.dump(data, json_file, indent=4)




"""
模型容器的异常处理
"""


import warnings


class TrainingError(Exception):
    """训练过程异常"""
    def __init__(self, msg, code=None):
        super().__init__(msg)
        self.code = code


class TrainingWarning(Warning):
    """训练过程警告"""
    pass



from hagike.basics.cv.file.image.im import *


def test_basics_cv_file(path: str = 'tmp/raw.png'):
    """basics.cv.file测试用例"""
    im = import_image(path, ImStyle.im_file)
    im.to_style_(ImStyle.im_ndarray)
    im.to_gray_()
    im.to_scale_(ImScale.float_n)
    im.to_scale_(ImScale.uint8_t)
    im.show_image(ImShow.windows)











"""
***常量表***
"""


from hagike.utils.enum import *


@advanced_enum()
class DataSrc(SuperEnum):
    """数据源类型"""
    mem = None
    """内存类"""
    file = None
    """pkl文件类"""
    compressed = None
    """压缩类"""
    else__ = None
    """其它类"""


@advanced_enum()
class DataForm(SuperEnum):
    """数据格式类型"""
    direct = None
    """直接可用"""
    mem_cached = None
    """内存缓存"""
    else__ = None
    """其它类"""


@advanced_enum()
class DataCompress(SuperEnum):
    """
    数据压缩类型 \n
    """
    else__ = None
    """其它类"""


@advanced_enum()
class DataInfo(SuperEnum):
    """
    数据包自定义数据类型
    """
    load_func = None
    """自定义加载函数"""
    load_check = None
    """自定义检查函数"""
    form_func = None
    """自定义格式转换函数"""
    mem_cacher = None
    """内存缓存器，如果希望从内存缓存读取数据，则需要指定"""
    compressed_type = None
    """压缩类型"""
    compressed_func = None
    """压缩函数"""
    others = dict()
    """用户自定义项，是一个字典"""



"""
***异常处理***
"""


import warnings


class DataPackError(Exception):
    """数据包错误"""
    def __init__(self, msg, code=None):
        super().__init__(msg)
        self.code = code


class DataPackWarning(Warning):
    """数据包警告"""
    pass



"""
***图像导入库与格式转换库*** \n

**description** \n
与**图像处理相关的库**主要有2个，现在阐明其区别与联系： \n
1. `cv2`，图像格式为 `np.ndarray`，作为向量格式，主要用于计算机视觉中对于图像的高级处理算法 \n
2. `Image`，图像格式为 `ImageFile` 或 `Image`，作为文件格式，主要用于对图像的直接操作，如格式转换等；但可以转化为`np.ndarray` \n
而**图片**在各层级上存在如下区别与联系： \n
1. 屏幕显示：坐标原点在左上角，y轴向下，x轴向右 \n
2. `np.ndarray`： \n
索引访问与形状获取：先y后x；np与cv2库函数调用，先x后y \n
- `im_ndarray`：源于Image的导入后直接转换；彩图为三维，$[H, W, C]$，通道顺序为 `R, G, B` ；灰度图为二维，$[H, W]$； \n
- `cv_ndarray`：源于cv2的导入，彩图的通道顺序为 `B, G, R` \n
3. `torch.Tensor`：卷积神经网络的期待输入为$[C, H, W]$，即使是灰度图也依然需要通道维度 \n
值得**注意**的是，python中的对整型的除法 `/` 默认会将原类型转换为小数格式，这对 `np.ndarray` 和 `torch.Tensor` 同样均成立； \n
如果希望整除，这需要使用 `//` 操作 \n
下面在各个图像操作函数内会具体说明操作差异。 \n

.. important::
    图像标准化容器 `ImStd` 设计的思想是，实现常量表中列出的最常用的配置； \n
    但对于非标准化的或待实现的配置，也允许使用，但要主动标记为 `else__` 来回避内置检查，但是不能调用该配置项相关的库函数，否则会报错； \n
    `auto__` 情况下在自动判断时会将不可识别项标记为 `else__` 并给出警告，
    考虑到自动判断是可能是不可靠的，最好主动标记为 `else__`，以避免不可预知的行为； \n
    对 `color` 和 `scale` 的判断依赖于 `style`，如果 `style` 项被标记为 `else__`，则 `color` 和 `scale` 也会被判断为 `else__`，
    但 `color` 与 `scale` 相互独立； \n

**term** \n
    im - 图像容器
    image - 图像文件
    style - 格式类型
    color - 颜色类型
    scale - 数据范围
    show - 显示方式

.. todo::
    1. 对用户添加 `check` 类函数来帮助用户确保数据符合预期
"""

from __future__ import annotations

import numpy as np

from .const import *
from .error import *
from datetime import datetime
from copy import deepcopy
from types import MappingProxyType
import cv2
import matplotlib.pyplot as plt
import warnings


def im_ap2ri(image: im_ndarray) -> im_ndarray:
    """`幅频、相频` 转 `实部、虚部`，目前仅支持 `im_ndarray` 格式，要求输入的通道数为2"""
    r_part = image[:, :, 0] * np.cos(image[:, :, 1])
    i_part = image[:, :, 0] * np.sin(image[:, :, 1])
    image = r_part + 1j * i_part
    return image


def im_ri2ap(image: im_ndarray) -> im_ndarray:
    """`实部、虚部` 转 `幅频、相频`，目前仅支持 `im_ndarray` 格式，要求输入的类型为复数类型"""
    magnitude = np.abs(image)
    phase = np.angle(image)
    image = np.stack((magnitude, phase), axis=-1)
    return image


class ImStd:
    """标准图像容器"""

    _np_int = (np.int64, np.int32, np.int16, np.int8,
               np.uint64, np.uint32, np.uint16, np.uint8, np.bool)
    """ `np` 中的整型集合"""
    _torch_int = (torch.int64, torch.int32, torch.int16, torch.int8,
                  torch.uint64, torch.uint32, torch.uint16, torch.uint8, torch.bool)
    """ `torch` 中的整型集合"""
    _np_float = (np.float64, np.float32, np.float16, np.complex64, np.complex128)
    """ `np` 中的浮点型集合"""
    _torch_float = (torch.float64, torch.float32, torch.float16, torch.complex64, torch.complex128)
    """ `torch` 中的浮点型集合"""
    _else = {ImTerm.style: ImStyle.else__, ImTerm.color: ImColor.else__, ImTerm.scale: ImScale.else__}
    """ 非标准化项在各 `term` 下对应的 `uuid` 值 """

    def __init__(self,
                 image: im_all,
                 style: uuid_t = ImStyle.auto__,
                 color: uuid_t = ImColor.auto__,
                 scale: uuid_t = ImScale.auto__,
                 check_style: bool = False,
                 check_color: bool = False,
                 check_scale: bool = False,
                 ) -> None:
        """
        初始化，`is_check` 指定是否进行检查与判断； \n
        如果未显式指定，且 `style` 或 `color` 或 `scale` 为 `_auto` 时使用懒惰思路，即当有转换需求的时候进行判断 \n
        """
        self._image: im_all = image
        self._style: uuid_t = style
        self._color: uuid_t = color
        self._scale: uuid_t = scale
        self._info: Dict = ImInfo.dict_()

        ImStyle.check_in_(self._style, all_or_index=True), ImColor.check_in_(self._color, all_or_index=True)
        self._is_check_style, self._is_check_color, self._is_check_scale = False, False, False
        if check_style:
            self._judge_style()
        if check_color:
            self._judge_color()
        if check_scale:
            self._judge_scale()

    @property
    def image(self) -> im_all | Any:
        """返回 `image` 本身"""
        return self._image

    @property
    def terms(self) -> Dict[uuid_t, Any]:
        """返回配置信息表"""
        return {ImTerm.style: self._style,
                ImTerm.color: self._color,
                ImTerm.scale: self._scale}

    @image.setter
    def image(self, image: Any) -> None:
        """重设 `image` 时更新检查状态，这里默认认为各配置项不变，若变则需要显式更改其它配置项，否则后续调用内置函数时自动检查会触发异常"""
        self._image = image
        self._is_check_style, self._is_check_color, self._is_check_scale = False, False, False

    @property
    def style(self) -> uuid_t:
        """给出 `style` 前进行检查"""
        self._judge_style()
        return self._style

    @style.setter
    def style(self, style: uuid_t) -> None:
        """重设 `style` 时更新检查状态"""
        self._style = style
        self._is_check_style = False

    @property
    def color(self) -> uuid_t:
        """给出 `color` 前进行检查"""
        self._judge_color()
        return self._color

    @color.setter
    def color(self, color) -> None:
        """重设 `color` 时更新检查状态，并重置 `ap_or_ri`"""
        self._color = color
        self._info[ImInfo.ap_or_ri] = None
        self._is_check_color = False

    @property
    def scale(self) -> uuid_t:
        """给出 `scale` 前进行检查"""
        self._judge_scale()
        return self._scale

    @scale.setter
    def scale(self, scale) -> None:
        """重设 `scale` 时更新检查状态，并重置 `vscale`"""
        self._scale = scale
        self._info[ImInfo.vscale] = None
        self._is_check_scale = False

    @property
    def info(self) -> MappingProxyType:
        """返回只读信息表"""
        return MappingProxyType(self._info)

    @property
    def others(self) -> Dict:
        """返回用户自定义信息表"""
        return self._info[ImInfo.others]

    @property
    def vscale(self) -> None | tuple:
        """返回自定义数据范围"""
        return self._info[ImInfo.vscale]

    @vscale.setter
    def vscale(self, vscale: vscale_t) -> None:
        """修改自定义数据范围，启用后 `scale` 自动变为 `else__`"""
        self.scale = ImScale.else__
        self._info[ImInfo.vscale] = vscale

    @property
    def ap_or_ri(self) -> None | bool:
        """返回频域类型，`幅度相位` / `实部虚部`"""
        self._judge_color()
        return self._info[ImInfo.ap_or_ri]

    @ap_or_ri.setter
    def ap_or_ri(self, ap_or_ri: bool) -> None:
        """修改频域类型"""
        self._is_check_color = False
        self._info[ImInfo.ap_or_ri] = ap_or_ri

    def _auto_style(self) -> None:
        """自动判断类型，如果值已经为 `else__`，则没有调用该函数的场景"""
        for uuid in ImStyle.iter_():
            if isinstance(self._image, ImStyle.get_value_(uuid)):
                self._style = uuid
                break
        else:
            self._style = ImStyle.else__
            warnings.warn(f"Image style {type(self._image)} is unknown!!!", ImageStyleWarning)

    def _judge_style(self) -> None:
        """检查图像类型"""
        if self._is_check_style:
            return
        self._is_check_style = True
        if self._style == ImStyle.auto__:
            self._auto_style()
        elif self._style == ImStyle.else__:
            # 由于 `color` 和 `scale` 依赖 `style`，故直接重设 `style`
            self._color = ImColor.else__
            self._scale = ImScale.else__
        else:
            if not isinstance(self._image, ImStyle.get_value_(self._style)):
                raise ImageStyleError(
                    f"Image style {type(self._image)} is not as given, {ImStyle.get_value_(self._style)}!!!")

    def _auto_color(self) -> None:
        """判断颜色类型，同时检查数据是否符合预期；如果值已经为 `else__`，则没有调用该函数的场景"""
        if self._style == ImStyle.im_file:
            if self._image.mode == 'L':
                self._color = ImColor.gray
            elif self._image.mode == 'RGB':
                self._color = ImColor.colored
            else:
                self._color = ImColor.else__
                warnings.warn(f"The image, {ImStyle.get_value_(self._style)}, "
                              f"has an unknown color format, f{self._image.mode}!!!", ImageInfoWarning)
        elif self._style == ImStyle.im_tensor:
            if len(self._image.shape) != 3:
                self._color = ImColor.else__
                warnings.warn(f"The image, {ImStyle.get_value_(self._style)}, "
                              f"has wrong shape, {self._image.shape}!!!", ImageInfoWarning)
            else:
                if self._image.shape[0] == 1:
                    if self._info[ImInfo.ap_or_ri] is None:
                        self._color = ImColor.gray
                    elif self._info[ImInfo.ap_or_ri] is False:
                        self._color = ImColor.freq
                    else:
                        self._color = ImColor.else__
                        warnings.warn(f"The image color is unrecognized, "
                                      f"with style {ImStyle.get_value_(self._style)}, "
                                      f"and shape {self._image.shape}!!!", ImageInfoWarning)
                elif self._image.shape[0] == 3:
                    self._color = ImColor.colored
                elif self._image.shape[0] == 2 and self._info[ImInfo.ap_or_ri] is True:
                    self._color = ImColor.freq
                else:
                    self._color = ImColor.else__
                    warnings.warn(f"The image, {ImStyle.get_value_(self._style)}, "
                                  f"has unknown numbers of color channel, {self._image.shape}!!!", ImageInfoWarning)
        elif self._style == ImStyle.im_ndarray or self._style == ImStyle.cv_ndarray:
            if len(self._image.shape) == 3:
                if self._image.shape[2] == 3:
                    self._color = ImColor.colored
                elif self._image.shape[2] == 2 and self._info[ImInfo.ap_or_ri] is True:
                    self._color = ImColor.freq
                else:
                    self._color = ImColor.else__
                    warnings.warn(f"The image color is unrecognized!!!", ImageInfoWarning)
            elif len(self._image.shape) == 2:
                if self._info[ImInfo.ap_or_ri] is None:
                    self._color = ImColor.gray
                elif self._info[ImInfo.ap_or_ri] is False:
                    self._color = ImColor.freq
                else:
                    self._color = ImColor.else__
                    warnings.warn(f"The image color is unrecognized!!!", ImageInfoWarning)
            else:
                self._color = ImColor.else__
                warnings.warn(f"The image, {ImStyle.get_value_(self._style)}, "
                              f"has unknown shapes, {self._image.shape}!!!", ImageInfoWarning)
        elif self._style == ImStyle.else__:
            self._color = ImColor.else__
        else:
            raise ImageStyleError(
                f"Image style, {ImStyle.get_value_(self._style)} is not implemented, a 'color'!!!")

    def _judge_color(self) -> None:
        """检查颜色类型"""
        if self._is_check_color:
            return
        self._is_check_color = True
        if self._color == ImColor.auto__:
            self._auto_color()
        elif self._color == ImColor.else__:
            pass
        else:
            color = self._color
            self._auto_color()
            if color != self._color:
                raise ImageInfoError(f"The image's color is different from the auto-detection!!!")

    def _auto_scale(self) -> None:
        """
        自动判断范围；如果值已经为 `else__`，则没有调用该函数的场景 \n
        如果未 `freq`，则 `scale` 不变，但此时不可用 `auto__`，因为此时 `scale` 指的是频谱来源的类型 \n
        """
        if self._color == ImColor.freq:
            if self._scale == ImScale.auto__:
                raise ImageInfoError(f"The scale cannot be `auto__` when the color is `freq`!!!")
            return
        if self._style == ImStyle.im_file:
            self._scale = ImScale.uint8_t
        elif self._style == ImStyle.im_tensor or self._style == ImStyle.cv_ndarray or self._style == ImStyle.im_ndarray:
            if self._image.dtype in self._np_int:
                self._scale = ImScale.uint8_t
            elif self._image.dtype in self._np_float:
                self._scale = ImScale.float_n
            else:
                self._scale = ImScale.else__
                warnings.warn(f"The image dtype, {self._image.dtype}, "
                              f"is not in {self._np_int} or {self._np_float}", ImageInfoWarning)
        elif self._style == ImStyle.else__:
            self._scale = ImScale.else__
        else:
            raise ImageStyleError(
                f"Image style {ImStyle.get_value_(self._style)} is not implemented, a 'scale'!!!")

    def _judge_scale(self) -> None:
        """检查范围"""
        if self._is_check_scale:
            return
        self._is_check_scale = True
        if self._scale == ImScale.auto__:
            self._auto_scale()
        elif self._scale == ImScale.else__:
            pass
        else:
            scale = self._scale
            self._auto_scale()
            if scale != self._scale:
                raise ImageInfoError(f"The image's scale is different from the auto-detection!!!")

    def _judge_all(self) -> None:
        """检查并刷新各项属性"""
        # `check_style` 需要在最前面，因为后面两项的检查依赖 `style`
        self._judge_style()
        self._judge_color()
        self._judge_scale()

    def _judge_else(self, *terms) -> None:
        """排除非标准化情形，输入任意数量的检查项"""
        ImTerm.check_include_(terms, all_or_index=False)
        status = self.terms
        for term in terms:
            if status[term] == self._else[term]:
                raise ImageElseError(f"{ImTerm.get_name_(term)} is an `else__` type, not implemented!!!")

    def _to_style(self, style: uuid_t) -> Tuple[im_all, Dict] | None:
        """
        `to` 函数的实际执行函数，若返回 `None` 则说明没有变化； \n
        """
        # 检查或判断
        ImStyle.check_in_(style, all_or_index=False)
        self._judge_all()
        self._judge_else(ImTerm.style)
        # 排除同类转换
        if style == self._style:
            return
        info = dict()
        # 转换为中间格式
        if self._style == ImStyle.im_ndarray:
            image = self._image
        elif self._style == ImStyle.cv_ndarray:
            if self._color == ImColor.colored:
                image = cv2.cvtColor(self._image, cv2.COLOR_BGR2RGB)
            else:
                image = self._image
        elif self._style == ImStyle.im_file:
            image = np.array(self._image)
        elif self._style == ImStyle.im_tensor:
            image = self._image.to('cpu')
            if self._color == ImColor.gray:
                image.squeeze_(0)
            elif self._color == ImColor.colored:
                image = image.permute(1, 2, 0)
            else:
                raise ImageInfoError(f"The image color format, {self._color}, is not implemented!!!")
            image = image.numpy()
        else:
            raise ImageStyleError(
                f"Image style {ImStyle.get_value_(self._style)} is not implemented a 'to' function!!!")
        # 从中间格式转换为目标格式并进行必要的转换检查
        if style == ImStyle.im_ndarray:
            pass
        elif style == ImStyle.cv_ndarray:
            if self._color == ImColor.colored:
                image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
        elif style == ImStyle.im_file:
            if image.dtype != np.uint8:
                warnings.warn(f"The image dtype, {image.dtype}, is not {np.uint8}", ImageFunctionWarning)
            if self._color == ImColor.colored:
                mode = 'RGB'
            elif self._color == ImColor.gray:
                mode = 'L'
            # 'freq' 颜色不可转为 'im_file' 格式
            else:
                raise ImageInfoError(f"The image color format, {self._color}, is not implemented!!!")
            image = Image.fromarray(image, mode=mode)
        elif style == ImStyle.im_tensor:
            image = torch.tensor(image, device='cpu')
            if self._color == ImColor.gray:
                image = np.expand_dims(image, axis=0)
            elif self._color == ImColor.colored:
                image = np.transpose(image, (2, 0, 1))
            else:
                raise ImageInfoError(f"The image color format, {self._color}, is not implemented!!!")
        else:
            raise ImageStyleError(
                f"Image style {ImStyle.get_value_(style)} is not implemented a 'to' function!!!")
        return image, info

    def to_style_(self, style: uuid_t) -> None:
        """
        格式转换为指定 `style` \n
        这里将 `im_ndarray` 作为中间格式用于转换 \n
        在因为格式原因无法进行转换时，会抛出异常 \n
        转换会就地进行 \n
        """
        result = self._to_style(style)
        if result is not None:
            image, info = result
            self._image = image
            self._style = style
            ImInfo.update_dict_(self._info, info)

    def to_style(self, style: uuid_t, is_new: bool = False) -> ImStd:
        """
        格式转换 \n
        `is_new=True`，则总是返回一个全新的 `ImStd` 图像容器，无论目标类型与原有类型是否一致 \n
        `is_new=False`，则仅在目标类型与原有类型不一致时返回深拷贝结果，否则返回本身； \n
        当明确知道后续只有只读操作后可以这样做，会节省算力与内存 \n
        要求 `style` 为标准化情形 \n
        """
        result = self._to_style(style)
        if result is None and is_new is False:
            return self
        else:
            im = deepcopy(self)
            if result is not None:
                image, info = result
                im._image = image
                ImInfo.update_dict_(im._info, info)
            im._style = style
            return im

    def _to_color(self, color: uuid_t) -> Tuple[im_all, Dict] | None:
        """
        `to_color` 的实质执行函数，如果本身就是灰度图则返回 `None` \n
         """
        self._judge_all()
        self._judge_else(ImTerm.style, ImTerm.color)
        if self._color == color:
            return
        info = dict()
        if self._color == ImColor.colored:
            if color == ImColor.gray:
                if self._style == ImStyle.im_file:
                    image = self._image.convert('L')
                elif self._style == ImStyle.im_ndarray:
                    image = cv2.cvtColor(self._image, cv2.COLOR_RGB2GRAY)
                elif self._style == ImStyle.cv_ndarray:
                    image = cv2.cvtColor(self._image, cv2.COLOR_BGR2GRAY)
                elif self._style == ImStyle.im_tensor:
                    # 在此处手动转化时需要注意前后数据类型不变
                    image = (self._image[0: 1, :, :] * 0.299 +
                             self._image[1: 2, :, :] * 0.587 +
                             self._image[2: 3, :, :] * 0.114).to(self._image.dtype)
                else:
                    raise ImageStyleError(
                        f"Image style {ImStyle.get_value_(self._style)} is not implemented from 'colored' function!!!")
            else:
                raise ImageInfoError(f"The image color format, {ImColor.get_name_(self._color)}, is illegal, "
                                     f"when transferring to {ImColor.get_name_(color)}!!!")
        elif self._color == ImColor.gray:
            if color == ImColor.colored:
                raise ImageInfoError(f"It's illegal to transfer color from 'gray' to 'colored'!!!")
            elif color == ImColor.freq:
                if self._style == ImStyle.im_ndarray or self._style == ImStyle.cv_ndarray:
                    image = np.fft.fft2(self._image)
                    image = np.fft.fftshift(image)
                else:
                    im = self.to_style(ImStyle.im_ndarray, is_new=True)
                    im._image = np.fft.fft2(im._image)
                    im._image = np.fft.fftshift(im._image)
                    im._color = ImColor.freq
                    im.to_style_(self._style)
                    image = im._image
                info[ImInfo.ap_or_ri] = False
            else:
                raise ImageInfoError(f"It's illegal to transfer color from 'gray' to 'colored'!!!")
        elif self._color == ImColor.freq:
            if color == ImColor.colored:
                raise ImageInfoError(f"It's illegal to transfer color from 'freq' to 'colored'!!!")
            elif color == ImColor.gray:
                im = self.to_style(ImStyle.im_ndarray, is_new=True)
                if self._info[ImInfo.ap_or_ri]:
                    im._image = im_ap2ri(im._image)
                src_scale = self._get_vscale(self._scale, self._info[ImInfo.vscale])
                im._image = np.fft.ifftshift(im._image)
                im._image = np.fft.ifft2(im._image).real.astype(src_scale[1])
                im.to_style_(self._style)
                image = im._image
                info[ImInfo.ap_or_ri] = False
            else:
                raise ImageInfoError(
                    f"Image color {ImColor.get_name_(color)} is not implemented from 'freq' function!!!")
        else:
            raise ImageInfoError(f"The image color format, {self._color}, is not implemented!!!")
        return image, info

    def to_color_(self, color: uuid_t) -> None:
        """
        转换颜色模式，并返回新的图像，本地转换 \n
        """
        result = self._to_color(color)
        if result is not None:
            image, info = result
            self._image = image
            self._color = color
            ImInfo.update_dict_(self._info, info)

    def to_color(self, color: uuid_t, is_new: bool = False) -> ImStd:
        """
        转换颜色模式 \n
        `is_new=True`，则总是返回一个全新的 `ImStd` 图像容器，无论原有是否一致 \n
        `is_new=False`，则仅在与原有不一致时返回深拷贝结果，否则返回本身； \n
        当明确知道后续只有只读操作后可以这样做，会节省算力与内存 \n
        要求 `style` 和 `color` 为标准化情形，否则报错 \n
        """
        result = self._to_color(color)
        if result is None and is_new is False:
            return self
        else:
            im = deepcopy(self)
            if result is not None:
                image, info = result
                im._image = image
                ImInfo.update_dict_(im._info, info)
            im._color = color
            return im

    @staticmethod
    def _get_vscale(scale: uuid_t, vscale: vscale_t | None = None) -> vscale_t:
        """获取 `vscale` 值"""
        if scale == ImScale.else__:
            if vscale is None:
                raise ImageInfoError(f"The dst vscale, is not set when scale is `else__`!!!")
        else:
            if vscale is not None:
                raise ImageInfoError(f"The dst vscale, {vscale}, is set "
                                     f"when scale, {vscale}, has been set in default!!!")
            vscale = ImScale.get_value_(scale)
        return vscale

    def _to_scale(self, scale: uuid_t, vscale: vscale_t | None = None) -> Tuple[im_all, Dict] | None:
        """
        `to_scale` 的实质执行函数，如果转换前后无变化则返回 `None` \n
        """
        ImScale.check_in_(scale, all_or_index=False)
        self._judge_all()
        self._judge_else(ImTerm.style)
        src_scale = self._get_vscale(self._scale, self._info[ImInfo.vscale])
        dst_scale = self._get_vscale(scale, vscale)
        if src_scale is None or dst_scale is None:
            raise ImageInfoError(f"src_scale {src_scale} or dst_scale {dst_scale} is None, "
                                 f"cannot implement 'to-scale'!!!")
        if src_scale == dst_scale:
            return
        info = dict()
        info[ImInfo.vscale] = vscale
        if self._style == ImStyle.im_file:
            im = self.to_style(ImStyle.im_ndarray)
            im._image = np.clip(
                (dst_scale[0][1] - dst_scale[0][0]) / (src_scale[0][1] - src_scale[0][0])
                * (im.image - src_scale[0][0]) + dst_scale[0][0],
                0, 255).astype(np.uint8)
            im.to_style_(ImStyle.im_file)
            image = im.image
        elif self._style == ImStyle.im_ndarray or self._style == ImStyle.cv_ndarray or self._style == ImStyle.im_tensor:
            image = ((dst_scale[0][1] - dst_scale[0][0]) / (src_scale[0][1] - src_scale[0][0])
                     * (self._image - src_scale[0][0]) + dst_scale[0][0])
            # `freq` 颜色不进行截断和类型转换
            if self._color != ImColor.freq:
                if self._style == ImStyle.im_ndarray or self._style == ImStyle.cv_ndarray:
                    image = np.clip(image, dst_scale[0][0], dst_scale[0][1]).astype(dst_scale[1])
                else:
                    # 将 `dtype` 类型由 `np.dtype` 映射为 `torch.dtype`
                    if dst_scale[1] in self._np_int:
                        dtype = self._torch_int[self._np_int.index(dst_scale[1])]
                    elif dst_scale[1] in self._np_float:
                        dtype = self._torch_float[self._np_float.index(dst_scale[1])]
                    else:
                        raise ImageInfoError(
                            f"The dst dtype, {dst_scale[1]}, is not in {self._np_int} or {self._np_float}")
                    image = torch.clip(image, dst_scale[0][0], dst_scale[0][1]).type(dtype)
        else:
            raise ImageStyleError(
                f"Image style {ImStyle.get_value_(self._style)} is not implemented a 'to_scale' function!!!")
        return image, info

    def to_scale_(self, scale: uuid_t, vscale: vscale_t | None = None) -> None:
        """本地转换"""
        result = self._to_scale(scale, vscale)
        if result is not None:
            image, info = result
            self._image = image
            self._scale = scale
            ImInfo.update_dict_(self._info, info)

    def to_scale(self, scale: uuid_t, vscale: vscale_t | None = None, is_new: bool = False) -> ImStd:
        """
        范围转换，如果源范围是非标准化的，则要求 `info.vscale` 被设置 \n
        `is_new=True`，则总是返回一个全新的 `ImStd` 图像容器 \n
        `is_new=False`，则仅在原有范围不为 `scale` 或 `vscale` 时返回深拷贝结果，否则返回本身； \n
        要求 `style` 为标准化情形，否则报错； \n
        若 `style` 为 `im_file`，则无视类型说明，转换后的实际类型总是 `np.uint8`，但设置类型时依然设置为目标类型 \n
        如果只是想要扩展 int 或 float 类型的精度，则直接用 image 替换原有 image 即可，均与默认选项兼容 \n
        .. important::
            `vscale` 只能在非标准情况下设置，如果标准情况下设置则会报错 \n
        """
        result = self._to_scale(scale, vscale)
        if result is None and is_new is False:
            return self
        else:
            im = deepcopy(self)
            if result is not None:
                image, info = result
                im._image = image
                ImInfo.update_dict_(im._info, info)
            im._scale = scale
            return im

    @staticmethod
    def _checkout_freq(im: ImStd, mag_or_phase: bool | None) -> ImStd:
        """将频域图转换为灰度图，用于显示的灰度图，转换后的灰度图范围总会被缩放至 $[0, 1]$ 区间"""
        if im._color == ImColor.freq:
            style = im.style
            im = im.to_style(ImStyle.im_ndarray, is_new=True)
            image_n = im.image
            if im._info[ImInfo.ap_or_ri] is False:
                image_n = im_ri2ap(image_n)
            if mag_or_phase is True:
                image_n = image_n[:, :, 0]
                image_n = np.log(image_n + 1)
            elif mag_or_phase is False:
                image_n = image_n[:, :, 1]
            else:
                raise ImageFunctionError(f"mag_or_phase is not set correctly when showing freq image!!!")
            image_n = image_n / np.max(image_n)
            im._image = image_n
            im._color = ImColor.gray
            im._scale = ImScale.float_n
            im.to_style_(style)
        return im

    def show_image(self, show: uuid_t, para=None) -> None:
        """
        显示图像，`show` 指定了图像显示的方式 \n
        `para` 中是附加参数项
        """
        ImShow.check_in_(show, all_or_index=False)
        self._judge_all()
        self._judge_else(ImTerm.style, ImTerm.color)
        para = ImOutput.dict_(para)

        if show == ImShow.inner:
            im = self.to_style(ImStyle.im_ndarray, is_new=False)
            im = self._checkout_freq(im, para[ImOutput.mag_or_phase])
            # 此处不会缩放至 `uint8`，因为 `plt.imshow` 本身就是自带缩放的
            if 'cmap' not in para[ImOutput.others].keys():
                if im._color == ImColor.gray or im._color == ImColor.freq:
                    cmap = 'gray'
                elif im._color == ImColor.colored:
                    cmap = None
                else:
                    raise ImageInfoError(f"The image color format, {self._color}, is not implemented!!!")
            else:
                cmap = para[ImOutput.others].pop('cmap')
            plt.imshow(im.image, cmap=cmap, **para[ImOutput.others])
            plt.show()
        elif show == ImShow.system:
            im = self._checkout_freq(self, para[ImOutput.mag_or_phase])
            im = im.to_scale(ImScale.uint8_t)
            im.to_style(ImStyle.im_file)
            im.image.show(**para[ImOutput.others])
        elif show == ImShow.windows:
            im = self.to_style(ImStyle.cv_ndarray, is_new=False)
            im = self._checkout_freq(im, para[ImOutput.mag_or_phase])
            im = im.to_scale(ImScale.uint8_t)
            if 'winname' not in para[ImOutput.others].keys():
                winname = datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')
            else:
                winname = para[ImOutput.others].pop('winname')
            cv2.imshow(winname, im.image)
            cv2.waitKey(0)
            cv2.destroyWindow(winname)
        else:
            raise ImageFunctionError(
                f"Showing Type {ImShow.get_name_(show)} is not implemented!!!")
        return

    def save_image(self, path: str, is_cache: bool = False, para=None) -> None:
        """
        保存图像 \n
        `is_cache` 指定了是否使用数据的序列化缓存格式，缓存格式下不需要 `style` 是标准化的 \n
        `para` 中是附加参数项 \n
        `freq` 格式只有在缓存模式下才是可被还原为频谱，否则只能作为灰度图 \n
        """
        if is_cache:
            save_data_to_pkl(self, path)
        else:
            self._judge_all()
            self._judge_else(ImTerm.style, ImTerm.color)
            para = ImOutput.dict_(para)
            ensure_path_writable(path, is_raise=True)
            im = self._checkout_freq(self, para[ImOutput.mag_or_phase])
            im = im.to_scale(ImScale.uint8_t)
            im = im.to_style(ImStyle.im_file, is_new=False)
            im.image.save(path, **para[ImOutput.others])



from hagike.models.train import *


def test_models_train_trainer():
    """测试models.train.trainer"""

    import torchvision.models as models
    import torchvision.transforms as transforms
    from hagike.models.train.dataset.tcv import TcvDataset, datasets
    from torch.optim import Adam
    from hagike.models.temp import ModuleTemp, ModuleKey
    from hagike.models.temp.common import ModuleTemp_MaskHead
    import torch.nn as nn
    import torch

    # 定义数据
    transform = transforms.Compose([transforms.ToTensor()])
    tcv_dataset = TcvDataset(
        dataset=datasets.CIFAR10,
        info={DatasetInfo.para: {'root': 'data/torchvision', 'download': True, 'transform': transform},
              DatasetInfo.sub_para: {'num_classes': 10, 'cache_size': 60000}}
    )
    loader = DataLoaderTemp(tcv_dataset, info={
        DataLoaderInfo.para: {'batch_size': 32, 'shuffle': True}
    })

    # 定义模型、优化器
    model = ModuleTemp_MaskHead(nodes={
        ModuleKey.bone: ModuleNode(models.resnet18(num_classes=10)),
        ModuleKey.head: ModuleNode(nn.Softmax(dim=-1))
    })

    # 定义其它
    optim = OptimTemp(info={OptimInfo.op_type: Adam, OptimInfo.para: {}, OptimInfo.lr: 0.001})
    criterion = CriterionTemp(info={CriterionInfo.crt_type: torch.nn.CrossEntropyLoss, CriterionInfo.para: {}})
    monitor = TrainerMonitor(info={MonitorInfo.logdir: 'tmp/logs', MonitorInfo.autostart: True})

    # 定义评估方式
    evaluator = EvaluatorTemp(info={EvaluatorInfo.effector_type: EffectorTemp, EvaluatorInfo.para: {}})

    # 定义训练器
    trainer = TrainerTemp(
        model, optim, criterion, loader, monitor, evaluator,
        info={TrainerInfo.device: 'cuda', TrainerInfo.max_epochs: 30}
    )
    trainer.train()
    trainer.end()




"""
***高级Enum类*** \n
**term**: \n
    SuperEnum - 高级枚举类的父类模板
    advanced_enum - 根级的装饰器，用于自动生成枚举类的各项配置
    uuid - 唯一标识符
    index - 索引
    name - 名称
    value - 值
**notice**: \n
    对于枚举成员：
    1. 本身的值在定义时是value，在访问时是uuid；对于部分IDE的代码高亮，为消除格式不符警告，可以类型标注为uuid_t
    2. index不应该被外部访问，而仅作为迭代器的索引值
    3. 命名不能以'__'开头，否则会被忽略
    4. 值如果是继承了SuperEnum的枚举类型，则会递归导入，需要确保此处枚举类的归属是唯一的，否则uuid会被多次修改；
    5. 如果未启用顺序访问索引，无法保证枚举类的书写顺序就是index顺序，顺序是由Python机制决定的（默认是按名称顺序）
    6. '_xxx'方式命名的枚举量为内置属性，不可用常规枚举量占用，且只能配置其中的配置字部分，否则会报错
    7. 其它的外显的普通枚举量可以命名为任意的'xxx'，前后不应该带有下划线!!!
    8. 对于'xxx__'方式命名的枚举量，这里会将其作为隐藏枚举量，不会被包含在index列表中，也不会分配index值
    9. 所有父类方法命名为'xxx_'，以此与隐藏枚举量进行区分
    10. 根类的值是会被包含在uuid及其映射中的，但由于其没有父类，因而不会被包含在任何index列表中，其index本身总为0
.. todo::
    待添加实例化常量表为本地配置表的功能； \n
    待添加从文件中加载 / 赋值常量表的功能； \n
    待添加对 `uuid_t` 类型的封装，使得封装内可以带有Enum类标识符，防止误输入属于另一个Enum类的标识符而绕过检查 \n
    待添加常量表类型的配置，如可选直接访问的是 `uuid` 还是 `value` \n
"""

from __future__ import annotations
from dataclasses import dataclass
from typing import Any, List, Tuple, Dict, Sequence, Set, Iterator, Mapping
from copy import deepcopy

# 重定义Enum中的标识符类型
uuid_t = int
index_t = int

_enum_conf_word = ('_value', '_sequence', '_hide', '_blank')
"""
SuperEnum类型配置字，在父类SuperEnum中定义的值为默认值 \n

:param _value:  \n
    对于group本身的赋值需要写在成员_value中，否则会被视为None，访问时依然通过value \n
:param _sequence:  \n
    如果在某子类下启用顺序访问索引，则需要赋值成员_sequence: Tuple[str]； \n
    其中需要按顺序列出所有成员名称；如果未列全或有不存在的成员名称则在初始化时会报错 \n
:param _hide: \n
    是否将类本身的值作为隐藏枚举值 \n
:param _blank:  \n
    打印时的单位空格长度 \n
"""

_enum_hide_word = ('_uuid', '_pack', '_length',
                   '_index2uuid', '_uuid2pack', '_uuid2sub', '_uuid2base',
                   '_uuid_all', '_uuid_hide', '_uuid_sub')
"""
SuperEnum类型隐藏字 \n

:param _uuid: \n
    子类本身的唯一标识符 \n
:param _pack: \n
    存储子类本身的信息，是信息的打包形式 \n
:param _length: \n
    子类的非隐藏成员数量，不包括子类本身 \n
:param _index2uuid:  \n
    子类的非隐藏成员索引到唯一标识符的映射，不包括子类本身 \n
:param _uuid2pack:  \n
    子类下的唯一标识符到数据包的映射，不包括子类本身 \n
:param _uuid2sub: \n
    子类下的唯一标识符到孙类的映射 \n
:param _uuid2base: \n
    根类下所有的唯一标识符到其父类的映射 \n
:param _uuid_all: \n
    子类下所有唯一标识符的集合 \n
:param _uuid_hide: \n
    子类下所有隐式枚举成员的集合 \n
:param _uuid_sub: \n
    子类下所有孙类枚举成员的集合 \n
"""


class EnumOccupiedError(Exception):
    """枚举类关键字占用异常"""
    def __init__(self, msg, code=None):
        super().__init__(msg)
        self.code = code


class EnumSequenceError(Exception):
    """枚举类顺序访问索引异常"""
    def __init__(self, msg, code=None):
        super().__init__(msg)
        self.code = code


class EnumTypeError(Exception):
    """枚举类的配置项的类型不正确"""
    def __init__(self, msg, code=None):
        super().__init__(msg)
        self.code = code


class EnumUuidError(Exception):
    """枚举类的uuid不存在"""
    def __init__(self, msg, code=None):
        super().__init__(msg)
        self.code = code


@dataclass
class _EnumPack:
    """常量数据包"""
    uuid: uuid_t = None
    index: index_t = None
    name: str = None
    value: Any = None

    def print(self, is_value: bool = False) -> None:
        """打印值，is_value表示是否打印值"""
        print(f"{self.name}({self.index}) -> {self.uuid}", end='')
        if is_value:
            print(f": {self.value}", end='')
        print()


class SuperEnum:
    """Enum类的父类"""
    # 配置属性
    _value: Any | None = None
    _sequence: Tuple[str] = None
    _hide: bool = False
    _blank: int = 4
    # 隐藏属性
    _uuid: uuid_t
    _pack: _EnumPack
    _length: int
    _index2uuid: List[uuid_t]
    _uuid2pack: Dict[uuid_t, _EnumPack]
    _uuid2sub: Dict[uuid_t, SuperEnum]
    _uuid2base: Dict[uuid_t, SuperEnum]
    _uuid_all: Set[uuid_t]
    _uuid_hide: Set[uuid_t]
    _uuid_sub: Set[uuid_t]

    @classmethod
    def get_uuid_(cls) -> uuid_t:
        """获得类本身的uuid"""
        return cls._uuid

    @classmethod
    def get_cls_(cls, uuid: uuid_t) -> SuperEnum:
        """获得子类，这要求 `uuid` 必须对应某子类，否则会报错"""
        return cls._uuid2sub[uuid]

    @classmethod
    def get_base_(cls, uuid: uuid_t) -> SuperEnum:
        """获得常量所对应的父类"""
        return cls._uuid2base[uuid]

    @classmethod
    def get_name_(cls, uuid: uuid_t) -> str:
        """获得枚举量的名称"""
        return cls._uuid2pack[uuid].name

    @classmethod
    def get_index_(cls, uuid: uuid_t) -> index_t:
        """获得枚举量的名称"""
        return cls._uuid2pack[uuid].index

    @classmethod
    def get_value_(cls, src: index_t | uuid_t, index_or_uuid: bool = False) -> Any:
        """返回深拷贝赋值，index_or_uuid决定根据index还是uuid返回值"""
        uuid = cls._index2uuid[src] if index_or_uuid else src
        pack_n = cls._uuid2pack[uuid]
        return deepcopy(pack_n.value)

    @classmethod
    def check_in_(cls, uuid: uuid_t, all_or_index: bool = False, is_raise: bool = True) -> bool:
        """检查是否uuid是否被枚举类包含，all_or_hide指定是否仅考虑显变量，is_raise指定若不存在是否引发错误"""
        if all_or_index:
            is_in = True if uuid in cls._uuid_all else False
        else:
            is_in = True if uuid in cls._index2uuid else False
        if is_raise:
            if not is_in:
                raise EnumUuidError(f"ERROR: {uuid} is not in enum!!!")
        return is_in

    @classmethod
    def check_include_(cls, enum_list: Sequence[uuid_t], all_or_index: bool = False) -> bool:
        """检查列表内的uuid是否都包含在枚举类中，all_or_hide指定是否仅考虑显变量"""
        is_include = True
        for uuid in enum_list:
            if not cls.check_in_(uuid, all_or_index):
                is_include = False
                break
        return is_include

    @classmethod
    def dict_(cls, enum_dict: Mapping[uuid_t, Any] = None, is_force: bool = True) -> Dict[uuid_t, Any]:
        """
        填补Enum类中不在常量表部分的默认赋值，本地替换； \n
        如果选中is_force则会检查enum_dict中的key是否都在enum类的非隐藏部分中，若不满足则会报错； \n
        此项检查用于排除不正常输入的字典项，如隐藏的enum成员。 \n
        """
        if enum_dict is None:
            enum_dict = dict()
        uuid_dict = enum_dict.keys()
        for index in range(cls._length):
            uuid = cls._index2uuid[index]
            if uuid not in uuid_dict:
                enum_dict[uuid] = cls.get_value_(uuid, False)
        # 检查是否完全包含于
        if is_force:
            if cls._length != len(uuid_dict):
                raise EnumUuidError(
                    f"ERROR: dict(len={len(uuid_dict)}) is not included in enum(len={cls._length})!!!")
        return enum_dict

    @classmethod
    def update_dict_(cls, src_dict: Dict[uuid_t, Any], new_dict: Dict[uuid_t, Any] | None, is_force: bool = True) \
            -> Dict[uuid_t, Any]:
        """
        用新配置更新原配置，`is_force` 指定在发现不在表中的量时是否报错； \n
        对 `src_dict` 本地替换 \n
        """
        if new_dict is None:
            return src_dict
        for key, value in new_dict.items():
            cls.check_in_(key, is_raise=is_force)
            src_dict[key] = value
        return src_dict

    @classmethod
    def list_(cls, enum_dict: Mapping[uuid_t, Any] = None, is_default: bool = False) -> List[Any]:
        """
        将dict根据index顺序进行排序，要求所有enum_dict中的key都被包含于enum类的非隐藏部分； \n
        is_default决定是否补全默认值；
        """
        enum_list = []
        if enum_dict is None:
            enum_dict = dict()
        uuid_dict = enum_dict.keys()
        # 检查是否完全包含于
        for uuid in uuid_dict:
            if uuid not in cls._index2uuid:
                raise EnumUuidError(f"ERROR: dict is not included in enum for uuid({uuid})!!!")
        for index in range(cls._length):
            uuid = cls._index2uuid[index]
            if uuid in uuid_dict:
                enum_list.append(enum_dict[uuid])
            elif is_default and uuid not in uuid_dict:
                enum_list.append(cls.get_value_(uuid, False))

        return enum_list

    @classmethod
    def len_(cls) -> int:
        """返回index数量"""
        return cls._length

    @classmethod
    def iter_(cls) -> Iterator[uuid_t]:
        """返回index2uuid迭代器"""
        return iter(cls._index2uuid)

    @classmethod
    def print_pack_(cls, uuid: uuid_t, blank: int = 0, is_value: bool = False) -> None:
        """打印单个枚举量的信息"""
        blank_str = '' + ' ' * blank
        pack = cls._uuid2pack[uuid]
        print(blank_str, end='')
        pack.print(is_value)

    @classmethod
    def print_(cls, is_value: bool = False) -> None:
        """打印枚举类单级信息"""
        print()
        cls._pack.print(is_value)
        # 优先顺序打印非隐藏值
        for uuid in cls._uuid_all:
            cls.print_pack_(uuid, cls._blank, is_value)
        print()

    @classmethod
    def tree_(cls, is_value: bool = False):
        """以树形结构递归打印该枚举类信息"""

        def regress_enum(cls_n: Any, blank_n: int) -> None:
            """递归列举"""
            # 优先顺序打印非隐藏值
            for uuid_n in cls_n._uuid_all:
                cls_n.print_pack_(uuid_n, blank_n, is_value)
                if uuid_n in cls_n._uuid_sub:
                    regress_enum(cls_n._uuid2sub[uuid_n], blank_n + cls_n._blank)

        # 递归入口
        print()
        cls._pack.print(is_value)
        regress_enum(cls, cls._blank)
        print()


def advanced_enum():
    """
    该函数作为常量表的装饰器，自动建立映射，子类与子成员均视为常量，封装为常量类型，仅用于顶级Enum。
    """

    def decorator(cls):
        """装饰器，进行常量封装"""

        def check_key(keys: Sequence, all_or_hide: bool = True) -> None:
            """检查是否存在关键字冲突，all_or_hide指定全部检查还是仅检查隐藏属性"""
            for word in keys:
                if word in _enum_hide_word:
                    raise EnumOccupiedError(f"ERROR: {word} in enum_hide_word, change a Name!!!")
            if all_or_hide:
                for word in keys:
                    if word in _enum_conf_word:
                        raise EnumOccupiedError(f"ERROR: {word} in enum_conf_word, change a Name!!!")

        def check_conf(cls_n: Any) -> None:
            """检查枚举类的配置项的类型与值是否正确"""
            if not isinstance(cls_n._hide, bool):
                raise EnumTypeError(f"ERROR: _hide_ typeof {type(cls_n._hide)} but not bool!!!")
            if isinstance(cls_n._blank, int):
                if cls_n._blank < 0:
                    raise EnumTypeError(f"ERROR: _blank_({cls_n._blank}) < 0!!!")
            else:
                raise EnumTypeError(f"ERROR: _blank_ typeof {type(cls_n._blank)} but not int!!!")

        def regress_enum(uuid_n: uuid_t, cls_n: Any) -> uuid_t:
            """逐目录递归赋值uuid常量表，不会赋值顶级enum组"""
            uuid2pack_n: Dict[uuid_t, _EnumPack] = dict()
            index2uuid_n: List[uuid_t | None] = list()
            uuid_hide_n: Set[uuid_t] = set()
            uuid2sub_n: Dict[uuid_t, Any] = dict()
            index_n = 0
            # cls_n.__dict__仅会取在枚举类进行定义的变量，而dir(cls_n)会取包括SuperEnum类内的所有变量
            all_attrs_n = list(cls_n.__dict__.keys())
            all_attrs_n.reverse()  # 默认普通枚举量在前，子类在后

            # 检查是否存在关键字占用
            check_key(all_attrs_n, all_or_hide=False)
            # 检查配置字是否合法
            check_conf(cls_n)

            # 判断是否启用局部顺序映射表，如果启用则判断是否合法（是否恰好一致）并调换顺序
            seq_n, is_seq, seq_len = cls_n._sequence, False, None
            if seq_n is not None:
                is_seq, seq_len = True, len(seq_n)
                # 检查以确保_sequence_中没有关键字冲突
                check_key(seq_n, all_or_hide=True)
                index2uuid_n = [None for _ in range(seq_len)]

            for attr_n in all_attrs_n:
                # 排除魔法属性，'_value'在父级中设置，不在本级设置
                if attr_n.startswith('__'):
                    continue
                # 排除内置属性，并确保子类不存在自定义的内置属性
                elif attr_n.startswith('_'):
                    if attr_n not in _enum_conf_word:
                        raise EnumOccupiedError(f"ERROR: {attr_n} not in enum_conf_word, change a Name!!!")
                # 检查枚举类中的隐藏变量是否存在命名规范问题
                elif attr_n.endswith('_') and not attr_n.endswith('__'):
                    raise EnumOccupiedError(
                        f"ERROR: {attr_n} is should end with '__' but not '_', change a Name!!!")
                # 检查枚举类中的隐藏变量是否与父类变量存在命名冲突
                # 此处检查仅出于保险起见，如果此处报错，那就是enum库中SuperEnum方法的命名上出现了问题，因为命名空间应该已经完全划分开
                elif hasattr(cls_n.__base__, attr_n):
                    raise EnumOccupiedError(
                        f"ERROR: {attr_n} is conflicted with function in SuperEnum, change a Name!!!")
                # 检查
                else:
                    # 重置标志位
                    is_hide, is_sub = False, False
                    # 赋值枚举类型
                    val_n = getattr(cls_n, attr_n)
                    # 判断类型是否为子类
                    if isinstance(val_n, type) and issubclass(val_n, SuperEnum):
                        is_sub = True
                    # 递归并处理子类
                    if is_sub:
                        # 先递归
                        uuid_n = regress_enum(uuid_n, val_n)
                        # 赋值子级group属性
                        uuid2sub_n[uuid_n] = val_n
                        pack_n = _EnumPack(uuid=uuid_n, name=attr_n, value=val_n._value)
                        val_n._uuid, val_n._pack = uuid_n, pack_n
                    # 处理一般枚举成员
                    else:
                        pack_n = _EnumPack(uuid=uuid_n, name=attr_n, value=val_n)
                        setattr(cls_n, attr_n, uuid_n)
                    # 检查是否为隐藏枚举量并处理
                    if is_sub:
                        if val_n._hide:
                            is_hide = True
                    else:
                        if attr_n.endswith('__'):
                            is_hide = True
                    if is_hide:
                        uuid_hide_n.add(uuid_n)
                    # 如果为隐藏属性，则跳过配置index过程
                    if not is_hide:
                        # 赋值索引值，如果启用了顺序索引则填入对应位置，否则挂到最后
                        if is_seq:
                            try:
                                index = seq_n.index(attr_n)
                                index2uuid_n[index] = uuid_n
                            except ValueError:
                                raise EnumSequenceError(f"ERROR: '{attr_n}' is not in _sequence_!!!")
                        else:
                            index = index_n
                            index2uuid_n.append(uuid_n)
                        pack_n.index = index
                        # 刷新计数器
                        index_n += 1
                    else:
                        pack_n.index = None
                    uuid2pack_n[uuid_n] = pack_n
                    uuid_n += 1
            # 如果启用了顺序索引，则检查_sequence_是否全部被包含
            if is_seq:
                if index_n != seq_len:
                    raise EnumSequenceError(f"ERROR: index_n({index_n}) != _sequence_({seq_len})!!!")
            # 赋值本级group属性
            cls_n._index2uuid = index2uuid_n
            cls_n._length = index_n
            cls_n._uuid2pack = uuid2pack_n
            cls_n._uuid_hide = uuid_hide_n
            cls_n._uuid2sub = uuid2sub_n
            cls_n._uuid_all = set(index2uuid_n) | uuid_hide_n
            cls_n._uuid_sub = set(uuid2sub_n.keys())

            return uuid_n

        def back_regress(cls_n: Any) -> None:
            """递归进行反向映射"""
            uuid2base_n = dict()
            for uuid_n in cls_n._uuid_all:
                uuid2base_n[uuid_n] = cls_n
                if uuid_n in cls_n._uuid2sub:
                    cls_sub = cls_n._uuid2sub[uuid_n]
                    back_regress(cls_sub)
                    uuid2base_n.update(cls_sub._uuid2base)
            cls_n._uuid2base = uuid2base_n

        # 前向递归入口
        uuid = 0
        uuid = regress_enum(uuid, cls)
        # 赋值根目录本身的属性，本身一般仅用于占位，无实际意义
        cls._uuid = uuid
        cls._pack = _EnumPack(uuid=uuid, index=(None if cls._hide else 0), name=cls.__name__, value=cls._value)

        # 建立反向映射
        back_regress(cls)

        return cls

    return decorator



"""
***错误与警告***
"""


import warnings


class ImageStyleError(Exception):
    """图像类型与实际类型不符"""
    def __init__(self, msg, code=None):
        super().__init__(msg)
        self.code = code


class ImageStyleWarning(Warning):
    pass


class ImageInfoError(Exception):
    """图像的颜色空间不支持或数据不符合预期"""
    def __init__(self, msg, code=None):
        super().__init__(msg)
        self.code = code


class ImageInfoWarning(Warning):
    pass


class ImageElseError(Exception):
    """对非标准化情形使用内置函数"""
    def __init__(self, msg, code=None):
        super().__init__(msg)
        self.code = code


class ImageElseWarning(Warning):
    pass


class ImageFunctionError(Exception):
    """图像功能性错误"""
    def __init__(self, msg, code=None):
        super().__init__(msg)
        self.code = code


class ImageFunctionWarning(Warning):
    pass



"""
模型容器的常量表
"""


from hagike.utils.enum import *
import torch


@advanced_enum()
class ModuleKey(SuperEnum):
    """模块构成"""
    _sequence = (
        'pre', 'tail', 'bone', 'head', 'final'
    )
    all__ = None    # 指代整个模型
    pre = None      # 预处理，将数据从原始格式转换为张量格式
    tail = None     # 尾部，将数据规范化，如批量归一化、嵌入等，以便于骨干网处理
    bone = None     # 骨干网，进行特征提取等操作
    head = None     # 头部，根据需求构造输出层格式
    final = None    # 激活层，获取最终输出


@advanced_enum()
class ModuleMode(SuperEnum):
    """模块运行模式"""
    train = None
    """训练模式"""
    val = None
    """评估模式"""
    predict = None
    """预测模式"""
    else__ = None
    """其它模式"""


@advanced_enum()
class ModuleInfo(SuperEnum):
    """模块信息表"""
    dtype = torch.float32
    """模块数据类型"""
    device = 'cpu'
    """模块所在设备"""
    mode = ModuleMode.predict
    """训练模式 / 评估模式 / 预测模式"""
    others = dict()
    """其它信息"""




"""
***常量表***
"""


from hagike.utils import *
import numpy as np
import torch
from PIL import Image, ImageFile


im_file = Image.Image | ImageFile.ImageFile
cv_ndarray = np.ndarray
im_ndarray = np.ndarray
im_tensor = torch.Tensor
im_all = im_file | cv_ndarray | im_ndarray | im_tensor


@advanced_enum()
class ImStyle(SuperEnum):
    """图像格式，值为须设计为可用于 `isinstance` 的类型"""
    im_file: uuid_t = (Image.Image, ImageFile.ImageFile)
    """Image格式"""
    cv_ndarray: uuid_t = cv_ndarray
    """cv2格式"""
    im_ndarray: uuid_t = im_ndarray
    """Image转array格式"""
    im_tensor: uuid_t = im_tensor
    """torch格式"""
    auto__: uuid_t = None
    """自动判断"""
    else__: uuid_t = None
    """非标准格式"""


@advanced_enum()
class ImScale(SuperEnum):
    """
    图像值范围 \n
    .. important::
        数据类型需要是 `np` 系列的，否则检查时会报错，内部会实现其它类型与 `np` 类型的转换！！！
    """
    uint8_t: uuid_t = ((0, 255), np.uint8)
    """
    标准RGB值 \n
    .. important::
        这里将所有的整数类型型都归为这一范围，
        且不会检查其范围是否真的在该区间内，但默认该区间为对应范围！！！
    """
    float_n: uuid_t = ((0.0, 1.0), np.float32)
    """
    归一化的像素值 \n
    .. important::
        这里将所有的浮点类型都归为这一范围，
        且不会检查其范围是否真的在该区间内，但默认该区间为对应范围！！！
    """
    auto__: uuid_t = None
    """自动判断"""
    else__: uuid_t = None
    """非标准范围"""


@advanced_enum()
class ImColor(SuperEnum):
    """颜色类型"""
    gray: uuid_t = None
    """单色"""
    colored: uuid_t = None
    """彩色"""
    freq: uuid_t = None
    """
    频域，限定单色或指定彩色某通道向此类型转换 \n
    频谱类型无视数值类型转换，容器中的 `scale` 指示的是频谱图来源的范围，而非频谱图本身的范围 \n
    容器中为实部虚部还是幅度相位，在附加信息表中会进行记录 \n
    """
    auto__: uuid_t = None
    """自动判断"""
    else__: uuid_t = None
    """非标准颜色"""


@advanced_enum()
class ImShow(SuperEnum):
    """图片显示方式"""
    inner = None
    """
    ***调用IDE嵌入查看器*** \n
    `plt.imshow` 可以接收 `Image` 和 `np.ndarray` 两种格式的输入，主要是用于显示数据图而非原始图像的； \n
    在显示或转换时，如果类型不是uint8类型的，容器会给出警告，因为自动转换可能是不可靠的 \n
    但优势在于，Pycharm等IDE中有嵌入的查看器，无需打开额外窗口，且是非阻塞的；同时，图像会带有坐标轴 \n
    在使用上，如果输入的是灰度图，且希望以灰度方式显示，则需要指定参数cmap='gray'，
    否则会以其它颜色映射方式显示(默认为'viridis'，即从蓝色到绿色再到黄色的渐变) \n
    .. warning::
        `plt.imshow` 总是会根据 `np.ndarray` 或 `Image` 转换后得到的 `np.ndarray` 的 `min` 和 `max` 自动缩放数据到$[0, 255]$
    """
    system = None
    """
    ***调用系统图片查看器*** \n
    `Image.show` ，`Image` 格式自带方式，非阻塞，但要与系统默认应用，且要打开额外窗口，较为麻烦 \n
    .. warning::
        在某些IDE的运行环境中，运行环境与外界环境是隔离的，这时会出现警告信息，且窗口不会被打开，但是也不会中断运行 \n
        但如果直接在系统内使用python运行，则不会有问题
    """
    windows = None
    """
    ***调用窗口查看器*** \n
    `cv2.imshow` 会启动cv2的私有窗口，但是是阻塞的，直到进行特定按键窗口才会被销毁 \n
    """


@advanced_enum()
class ImTerm(SuperEnum):
    """`ImStd` 中的项"""
    image__ = None
    style = None
    color = None
    scale = None
    info__ = None


@advanced_enum()
class ImInfo(SuperEnum):
    """
    图片的个性化信息，主要服务于非标准化情况，并提供用户个性化存储的容器 \n
    这里列出的仅为常用类型，实际使用时包括且不限于此，这里仅 others 中的项是用户有自由设置权限的，其它都需要通过接口设置 \n
    """
    vscale = None
    """范围，要求为有两项的元组"""
    ap_or_ri = None
    """频域情况下，图像为 `amplitude-phase` 还是 `real-imaginary`"""
    others = dict()
    """用户自定义项，是一个字典"""


@advanced_enum()
class ImOutput(SuperEnum):
    """
    显示图像时的选择性参数 \n
    """
    mag_or_phase = None
    """如果是频域，那显示幅频还是相频"""
    others = dict()
    """用户自定义项，是一个字典，会在下层调用时展开"""


vscale_t = Tuple[Tuple[int | float, int | float], np.dtype]
"""数据范围类型"""



"""
***日志容器*** \n
这里的日志器希望实现以下功能： \n
1. 根据日志系统配置控制终端的消息输出 \n
2. 记录日志内容 \n
3. 提供日志统计与分析功能 \n
4. 提供保存与读取日志的方法 \n
5. 可以接入其它日志库，提供转换方法 \n
6. 可以进行日志合并等高级操作 \n
7. 可以提供跨进程的访问机制（可以作为子类拓展实现） \n
8. 定义基准消息格式 \n
9. 提供全局默认日志器 \n
.. todo::
    未全部完成
"""

from typing import Dict, Any
from .const import uuid_t, LoggerConf, LoggerStatus
from dataclasses import dataclass
from loguru import logger


@dataclass
class LogTemp:
    """日志项"""
    # todo: 实现对如pbar等回滚的日志项支持
    uuid: int = None
    """日志标识符"""
    time: int | float = None
    """时间"""
    level: uuid_t = None
    """级别"""
    src: str = None
    """来源"""
    event_main: uuid_t = None
    """主事件号"""
    event_sub: uuid_t = None
    """次事件号"""
    msg: str = None
    """消息"""
    else__: Any = None
    """其它附加要素"""


class LoggerTemp:
    """
    日志管理器模板 \n
    设计思路与实现目标： \n
    1. 为日志提供全局配置方法，使得可以控制日志记录与输出行为 \n
    2. 提供格式化的日志保存方法 \n
    3. 消息是日志的一种子形式，可以被包含在日志项范围内 \n
    """

    def __init__(self, conf: Dict[uuid_t, Any] | None = None, is_init: bool = True) -> None:
        """根据配置初始化日志模板"""
        self._conf = LoggerConf.dict_(conf)
        self._record = list()
        self._uuid = 0
        self._init_time = 0
        self._status = LoggerStatus.inited
        # todo: 基于配置的进一步初始化
        if is_init:
            self.init()

    def init(self) -> None:
        """启动日志管理器"""
        if self._status != LoggerStatus.inited:
            return
        # todo: 配置时间，写入日志头等
        self._status = LoggerStatus.running

    def end(self) -> None:
        """结束日志管理器"""
        # todo: 释放资源，写入日志尾等
        self._status = LoggerStatus.ended

    @property
    def status(self): return self._status

    def add_log(self, log: LogTemp, is_print: bool = True):
        """添加日志项"""
        # todo: 实现功能、过滤等
        if is_print:
            self.print_log(log)

    def print_msg(self, msg: str):
        """配置并打印消息，但不进行日志记录"""
        # todo: 实现可配置打印
        logger.info(msg)

    def print_log(self, log: LogTemp):
        """打印日志项"""
        # todo: 实现可配置打印
        print(log.msg)


# 全局日志记录器
logger_g = LoggerTemp()



"""
***数据缓存类*** \n
"""


import os
import pickle
from typing import Any, Callable
from functools import lru_cache
from .file import *


def save_data_to_pkl(data: Any, path: str) -> None:
    """将数据缓存为pkl格式"""
    ensure_path_writable(path)
    dir_name = os.path.dirname(path)
    os.makedirs(dir_name, exist_ok=True)
    with open(path, 'wb') as f:
        pickle.dump(data, f)


def load_data_from_pkl(path: str) -> Any:
    """将数据从pkl加载"""
    check_path_readable(path)
    with open(path, 'rb') as f:
        data = pickle.load(f)
    return data


class MemCacher:
    """内存缓存器"""
    def __init__(self, func: Callable, max_size: int, typed: bool = False):
        self.func = func
        self.cached_func = lru_cache(maxsize=max_size, typed=typed)(self.func)

    def __call__(self, *args, **kwargs):
        return self.cached_func(*args, **kwargs)





from setuptools import setup, find_packages


def read(file) -> list:
    """读取 xxx.txt，并转换为列表"""
    try:
        with open(file, 'r') as f:
            requirements = f.read().splitlines()
        requirements = [item for item in requirements
                        if not item.startswith('matlabengineforpython')]
        return requirements
    except FileNotFoundError:
        return []


setup(
    name='hagike',
    version='0.0.7',
    packages=find_packages(),
    install_requires=read('requirements.txt'),
    include_package_data=True,
    package_data={
        'hagike': read('data.txt'),
    },
    description='hagike toolkit for everything',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='hagikehappy',
    author_email='hagikehappy@163.com',
    python_requires='>=3.10',
    classifiers=[  
        "Programming Language :: Python :: 3",  
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: OS Independent",  
    ],  
    project_urls={  
        'Homepage': 'https://github.com/hagikehappy/hagike',
        'Issues': 'https://github.com/hagikehappy/hagike/issues',
    },  
)




"""
***判断器***
"""


import torch.nn as nn
from .const import uuid_t, CriterionInfo
from ..temp import ModuleNode
from .error import TrainingError
from typing import Dict, Any


class CriterionTemp:
    """用于计算损失函数的容器"""

    def __init__(self, info: Dict[uuid_t, Any] | None = None) -> None:
        """创建评估器容器，需要先初始化才能使用"""
        self._info = CriterionInfo.dict_(info)
        self._crt = None
        self._is_init = False

    def init(self, model: ModuleNode | None = None) -> None:
        """初始化损失函数容器，可根据 `model` 动态生成评估器，但基准实现中不使用"""
        if self._is_init:
            return
        self._is_init = True
        self._crt = self._info[CriterionInfo.crt_type](**self._info[CriterionInfo.para])

    @property
    def crt(self) -> nn.Module:
        """返回评估器"""
        if self._is_init:
            return self._crt
        else:
            raise TrainingError(f"visiting crt when criterion uninitialized!!!")

    def __call__(self, *args, **kwargs):
        """重定向自身调用至 `self._crt`"""
        return self.crt(*args, **kwargs)

    def __getattr__(self, name: str):
        """重定向属性或方法至 `self._crt`"""
        return getattr(self.crt, name)




"""
hagike套件，由hagikehappy制作 \n
`Author`: `hagikehappy` \n
`Institute`: `School of Electronic Science and Engineering, Nanjing University` \n
`Logo`: \n
<img src="__res__/hagikehappy.jpg" style="zoom:30%;" /> \n
"""




"""
***数据集***
"""


import torch
from torch.utils.data import Dataset
from typing import Any, Dict, Tuple
from ..const import uuid_t, DatasetInfo
from ..error import TrainingError
from ...temp.const import ModuleInfo


class _DatasetTemp(Dataset):
    """二次封装Dataset，加入内部配置"""

    def __init__(self, dataset: Dataset, para: Dict[str, Any]) -> None:
        """初始化"""
        self._dataset = dataset
        self._para = para

    def __getitem__(self, index) -> Tuple[torch.Tensor, torch.Tensor]:
        """获取输入并转换设备"""
        inputs, labels = self._dataset[index]
        if isinstance(inputs, torch.Tensor):
            inputs = inputs.to(device=self._para[ModuleInfo.device], dtype=self._para[ModuleInfo.dtype])
        else:
            inputs = torch.tensor(inputs, device=self._para[ModuleInfo.device], dtype=self._para[ModuleInfo.dtype])
        if isinstance(labels, torch.Tensor):
            labels = labels.to(device=self._para[ModuleInfo.device], dtype=self._para[ModuleInfo.dtype])
        else:
            labels = torch.tensor(labels, device=self._para[ModuleInfo.device], dtype=self._para[ModuleInfo.dtype])
        return inputs, labels

    def __len__(self) -> int:
        """获取数据集数量"""
        return len(self._dataset)

    def __getattr__(self, name: str) -> Any:
        """重定向属性或方法至 `self._dataset`"""
        return getattr(self._dataset, name)


class DatasetTemp:
    """
    数据集容器 \n
    由于数据集的处理本身高度多变，因此这里仅给出接口而不包含基本实现
    """

    def __init__(self, train: Dataset | None = None,
                 val: Dataset | None = None,
                 info: Dict[uuid_t, Any] | None = None) -> None:
        """初始化"""
        self._info = DatasetInfo.dict_(info)
        # 基准实现中 train 和 val 直接赋值得到，未用到 self._info[DatasetInfo.para]；但继承类中可能用于创建 train 和 val
        self._train_dataset = train
        self._val_dataset = val
        self._container = _DatasetTemp
        self._train = None
        self._val = None
        self._is_init = False

    def init(self, model_info: Dict[uuid_t, Any]):
        """配置目标设备等"""
        if self._is_init:
            return
        self._is_init = True
        self._info[DatasetInfo.sub_para].update(model_info)
        self._train = self._container(self._train_dataset, self._info[DatasetInfo.sub_para])
        self._val = self._container(self._val_dataset, self._info[DatasetInfo.sub_para])

    @property
    def train(self) -> Dataset:
        if self._is_init:
            return self._train
        else:
            raise TrainingError(f"visiting dataset when DatasetTemp uninitialized!!!")

    @property
    def val(self) -> Dataset:
        if self._is_init:
            return self._val
        else:
            raise TrainingError(f"visiting dataset when DatasetTemp uninitialized!!!")



